#!/usr/bin/env python3
# coding: utf-8
"""
Created on Apr 2020
@author: Nathan de Lara <ndelara@enst.fr>
"""
from typing import Optional, Union

import numpy as np
from scipy import sparse
from scipy.spatial import cKDTree

from sknetwork.embedding.base import BaseEmbedding
from sknetwork.embedding.spectral import Spectral
from sknetwork.linalg import normalize
from sknetwork.utils.check import check_adjacency_vector, check_format, check_square, is_symmetric
from sknetwork.utils.format import directed2undirected


class Spring(BaseEmbedding):
    """Spring layout for displaying small graphs.

    * Graphs
    * Digraphs

    Parameters
    ----------
    n_components : int
        Dimension of the graph layout.
    strength : float
        Intensity of the force that moves the nodes.
    n_iter : int
        Number of iterations to update positions.
    tol : float
        Minimum relative change in positions to continue updating.
    approx_radius : float
        If a positive value is provided, only the nodes within this distance a given node are used to compute
        the repulsive force.
    position_init : str
        How to initialize the layout. If 'spectral', use Spectral embedding in dimension 2,
        otherwise, use random initialization.

    Attributes
    ----------
    embedding_ : np.ndarray
        Layout.

    Example
    -------
    >>> from sknetwork.embedding import Spring
    >>> from sknetwork.data import karate_club
    >>> spring = Spring()
    >>> adjacency = karate_club()
    >>> embedding = spring.fit_transform(adjacency)
    >>> embedding.shape
    (34, 2)

    Notes
    -----
    Simple implementation designed to display small graphs.

    References
    ----------
    Fruchterman, T. M. J.,  Reingold, E. M. (1991).
    `Graph Drawing by Force-Directed Placement.
    <https://onlinelibrary.wiley.com/doi/pdf/10.1002/spe.4380211102>`_
    Software – Practice & Experience.
    """
    def __init__(self, n_components: int = 2, strength: float = None, n_iter: int = 50, tol: float = 1e-4,
                 approx_radius: float = -1, position_init: str = 'random'):
        super(Spring, self).__init__()
        self.n_components = n_components
        self.strength = strength
        self.n_iter = n_iter
        self.tol = tol
        self.approx_radius = approx_radius
        if position_init not in ['random', 'spectral']:
            raise ValueError('Unknown initial position, try "spectral" or "random".')
        else:
            self.position_init = position_init

    def fit(self, adjacency: Union[sparse.csr_matrix, np.ndarray], position_init: Optional[np.ndarray] = None,
            n_iter: Optional[int] = None) -> 'Spring':
        """Compute layout.

        Parameters
        ----------
        adjacency :
            Adjacency matrix of the graph, treated as undirected.
        position_init : np.ndarray
            Custom initial positions of the nodes. Shape must be (n, 2).
            If ``None``, use the value of self.pos_init.
        n_iter : int
            Number of iterations to update positions.
            If ``None``, use the value of self.n_iter.

        Returns
        -------
        self: :class:`Spring`
        """
        adjacency = check_format(adjacency)
        check_square(adjacency)
        if not is_symmetric(adjacency):
            adjacency = directed2undirected(adjacency)
        n = adjacency.shape[0]

        position = np.zeros((n, self.n_components))
        if position_init is None:
            if self.position_init == 'random':
                position = np.random.randn(n, self.n_components)
            elif self.position_init == 'spectral':
                position = Spectral(n_components=self.n_components, normalized=False).fit_transform(adjacency)
        elif isinstance(position_init, np.ndarray):
            if position_init.shape == (n, self.n_components):
                position = position_init.copy()
            else:
                raise ValueError('Initial position has invalid shape.')
        else:
            raise TypeError('Initial position must be a numpy array.')

        if n_iter is None:
            n_iter = self.n_iter

        if self.strength is None:
            strength = np.sqrt((1 / n))
        else:
            strength = self.strength

        pos_max = position.max(axis=0)
        pos_min = position.min(axis=0)
        step_max: float = 0.1 * (pos_max - pos_min).max()
        step: float = step_max / (n_iter + 1)
        tree = None

        delta = np.zeros((n, self.n_components))
        for iteration in range(n_iter):
            delta *= 0
            if self.approx_radius > 0:
                tree = cKDTree(position)

            for i in range(n):
                # attraction
                indices = adjacency.indices[adjacency.indptr[i]:adjacency.indptr[i+1]]
                attraction = adjacency.data[adjacency.indptr[i]:adjacency.indptr[i+1]] / strength

                grad = position[i] - position[indices]
                attraction *= np.linalg.norm(grad, axis=1)
                attraction = (grad * attraction[:, np.newaxis]).sum(axis=0)

                # repulsion
                if tree is None:
                    grad: np.ndarray = (position[i] - position)  # shape (n, n_components)
                    distance: np.ndarray = np.linalg.norm(grad, axis=1)  # shape (n,)
                else:
                    neighbors = tree.query_ball_point(position[i], self.approx_radius)
                    grad: np.ndarray = (position[i] - position[neighbors])  # shape (n_neigh, n_components)
                    distance: np.ndarray = np.linalg.norm(grad, axis=1)  # shape (n_neigh,)

                distance = np.where(distance < 0.01, 0.01, distance)
                repulsion = (grad * (strength / distance)[:, np.newaxis] ** 2).sum(axis=0)

                # total force
                delta[i]: np.ndarray = repulsion - attraction

            length = np.linalg.norm(delta, axis=0)
            length = np.where(length < 0.01, 0.1, length)
            delta = delta * step_max / length
            position += delta
            step_max -= step
            err: float = np.linalg.norm(delta) / n
            if err < self.tol:
                break

        self.embedding_ = position
        return self

    def predict(self, adjacency_vectors: Union[sparse.csr_matrix, np.ndarray]) -> np.ndarray:
        """Predict the embedding of new rows, defined by their adjacency vectors.

        Parameters
        ----------
        adjacency_vectors :
            Adjacency vectors of nodes.
            Array of shape (n_col,) (single vector) or (n_vectors, n_col)

        Returns
        -------
        embedding_vectors : np.ndarray
            Embedding of the nodes.
        """
        self._check_fitted()
        embedding = self.embedding_
        n = embedding.shape[0]

        adjacency_vectors = check_adjacency_vector(adjacency_vectors, n)
        embedding_vectors = normalize(adjacency_vectors).dot(embedding)

        if embedding_vectors.shape[0] == 1:
            embedding_vectors = embedding_vectors.ravel()

        return embedding_vectors



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Apr 4, 2019
@author: Nathan de Lara <ndelara@enst.fr>
"""
import warnings
from typing import Union, Optional

import numpy as np
from scipy import sparse


def has_nonnegative_entries(entry: Union[sparse.csr_matrix, np.ndarray]) -> bool:
    """True if the array has non negative entries."""
    if type(entry) == sparse.csr_matrix:
        return np.all(entry.data >= 0)
    else:
        return np.all(entry >= 0)


def check_nonnegative(entry: Union[sparse.csr_matrix, np.ndarray]):
    """Check whether the array has non negative entries."""
    if not has_nonnegative_entries(entry):
        raise ValueError('Only nonnegative values are expected.')
    else:
        return


def has_positive_entries(entry: np.ndarray) -> bool:
    """True if the array has positive entries."""
    if type(entry) != np.ndarray:
        raise TypeError('Entry must be a dense NumPy array.')
    else:
        return np.all(entry > 0)


def check_positive(entry: Union[sparse.csr_matrix, np.ndarray]):
    """Check whether the array has positive entries."""
    if not has_positive_entries(entry):
        raise ValueError('Only positive values are expected.')
    else:
        return


def is_proba_array(entry: np.ndarray) -> bool:
    """True if each line of the array has non negative entries which sum to 1."""
    if len(entry.shape) == 1:
        return has_nonnegative_entries(entry) and np.isclose(entry.sum(), 1)
    elif len(entry.shape) == 2:
        n_row, n_col = entry.shape
        err = entry.dot(np.ones(n_col)) - np.ones(n_row)
        return has_nonnegative_entries(entry) and np.isclose(np.linalg.norm(err), 0)
    else:
        raise TypeError('Entry must be one or two-dimensional array.')


def is_square(adjacency: Union[sparse.csr_matrix, np.ndarray]) -> bool:
    """True if the matrix is square."""
    return adjacency.shape[0] == adjacency.shape[1]


def check_square(adjacency: Union[sparse.csr_matrix, np.ndarray]):
    """Check whether a matrix is square and return an error otherwise."""
    if is_square(adjacency):
        return
    else:
        raise ValueError('The adjacency is expected to be square.')


def is_symmetric(adjacency: sparse.csr_matrix, tol: float = 1e-10) -> bool:
    """True if the matrix is symmetric."""
    sym_error = adjacency - adjacency.T
    return np.all(np.abs(sym_error.data) <= tol * np.abs(adjacency.data.max()))


def check_symmetry(adjacency: sparse.csr_matrix, tol: float = 1e-10):
    """Check whether a matrix is symmetric and return an error otherwise."""
    if is_symmetric(adjacency, tol):
        return
    else:
        raise ValueError('The adjacency is expected to be symmetric.')


def is_connected(adjacency: sparse.csr_matrix) -> bool:
    """Check whether a graph is weakly connected.

    Parameters
    ----------
    adjacency:
        Adjacency matrix of the graph.
    """
    n_cc = sparse.csgraph.connected_components(adjacency, (not is_symmetric(adjacency)), 'weak', False)
    return n_cc == 1


def check_connected(adjacency: sparse.csr_matrix):
    """Check is a graph is connected and return an error otherwise."""
    if is_connected(adjacency):
        return
    else:
        raise ValueError('The adjacency is expected to be connected.')


def make_weights(distribution: str, adjacency: sparse.csr_matrix) -> np.ndarray:
    """Array of weights from a matrix and a desired distribution.

   Parameters
   ----------
   distribution:
       Distribution for node sampling. Only ``'degree'`` or ``'uniform'`` are accepted.
   adjacency:
       The adjacency matrix of the neighbors.

   Returns
   -------
   node_weights: np.ndarray
       Weights of nodes.
    """
    n = adjacency.shape[0]
    distribution = distribution.lower()
    if distribution == 'degree':
        node_weights_vec = adjacency.dot(np.ones(adjacency.shape[1]))
    elif distribution == 'uniform':
        node_weights_vec = np.ones(n)
    else:
        raise ValueError('Unknown distribution of node weights.')
    return node_weights_vec


def check_format(adjacency: Union[sparse.csr_matrix, np.ndarray]) -> sparse.csr_matrix:
    """Check whether the matrix is a NumPy array or a Scipy CSR matrix and return
    the corresponding Scipy CSR matrix.
    """
    if type(adjacency) not in {sparse.csr_matrix, np.ndarray}:
        raise TypeError('Adjacency must be in Scipy CSR format or Numpy ndarray format.')
    else:
        return sparse.csr_matrix(adjacency)


def check_is_proba(entry: Union[float, int], name: str = None):
    """Check whether the number is non-negative and less than or equal to 1."""
    if name is None:
        name = 'Probabilities'
    if type(entry) not in [float, int]:
        raise TypeError('{} must be floats (or ints if 0 or 1).'.format(name))
    if entry < 0 or entry > 1:
        raise ValueError('{} must have value between 0 and 1.'.format(name))


def check_damping_factor(damping_factor: float):
    """Check if the damping factor has a valid value."""
    if damping_factor < 0 or damping_factor >= 1:
        raise ValueError('A damping factor must have a value in [0, 1[.')


def check_weights(weights: Union['str', np.ndarray], adjacency: Union[sparse.csr_matrix, sparse.csc_matrix],
                  positive_entries: bool = False) -> np.ndarray:
    """Check whether the weights are a valid distribution for the adjacency and return a probability vector.

    Parameters
    ----------
    weights:
        Probabilities for node sampling in the null model. ``'degree'``, ``'uniform'`` or custom weights.
    adjacency:
        The adjacency matrix of the graph.
    positive_entries:
        If true, the weights must all be positive, if False, the weights must be non-negative.

    Returns
    -------
    node_weights: np.ndarray
        Valid weights of nodes.
    """
    n = adjacency.shape[0]
    if type(weights) == np.ndarray:
        if len(weights) != n:
            raise ValueError('The number of node weights must match the number of nodes.')
        else:
            node_weights_vec = weights
    elif type(weights) == str:
        node_weights_vec = make_weights(weights, adjacency)
    else:
        raise TypeError(
            'Node weights must be a known distribution ("degree" or "uniform" string) or a custom NumPy array.')

    if positive_entries and not has_positive_entries(node_weights_vec):
        raise ValueError('All weights must be positive.')
    else:
        if np.any(node_weights_vec < 0) or node_weights_vec.sum() <= 0:
            raise ValueError('Node weights must be non-negative with positive sum.')

    return node_weights_vec


def check_probs(weights: Union['str', np.ndarray], adjacency: Union[sparse.csr_matrix, sparse.csc_matrix],
                positive_entries: bool = False) -> np.ndarray:
    """Check whether the weights are a valid distribution for the adjacency
    and return a normalized probability vector.
    """
    weights = check_weights(weights, adjacency, positive_entries)
    return weights / np.sum(weights)


def check_random_state(random_state: Optional[Union[np.random.RandomState, int]]):
    """Check whether the argument is a seed or a NumPy random state. If None, numpy.random is used by default."""
    if random_state is None or random_state is np.random:
        return np.random
    elif type(random_state) == int:
        return np.random.RandomState(random_state)
    elif type(random_state) == np.random.RandomState:
        return random_state
    else:
        raise TypeError('To specify a random state, pass the seed (as an int) or a NumPy random state object.')


def check_seeds(seeds: Union[np.ndarray, dict], n: int) -> np.ndarray:
    """Check the format of seeds for semi-supervised algorithms."""

    if isinstance(seeds, np.ndarray):
        if len(seeds) != n:
            raise ValueError('Dimensions mismatch between adjacency and seeds vector.')
    elif isinstance(seeds, dict):
        keys, values = np.array(list(seeds.keys())), np.array(list(seeds.values()))
        if min(values) < 0:
            warnings.warn(Warning("Negative values will not be taken into account."))
        seeds = -np.ones(n)
        seeds[keys] = values
    else:
        raise TypeError('"seeds" must be a dictionary or a one-dimensional array.')
    return seeds


def check_n_neighbors(n_neighbors: int, n_seeds: int):
    """Set the number of neighbors so that it does not exceed the number of labeled samples."""
    if n_neighbors > n_seeds:
        warnings.warn(Warning("The number of neighbors cannot exceed the number of seeds. Changed accordingly."))
        n_neighbors = n_seeds
    return n_neighbors


def check_labels(labels: np.ndarray):
    """Check labels of the seeds for semi-supervised algorithms."""

    classes: np.ndarray = np.unique(labels[labels >= 0])
    n_classes: int = len(classes)

    if n_classes < 2:
        raise ValueError('There must be at least two distinct labels.')
    else:
        return classes, n_classes


def check_n_jobs(n_jobs: Optional[int] = None):
    """Parse the ``n_jobs`` parameter for multiprocessing."""
    if n_jobs == -1:
        return None
    elif n_jobs is None:
        return 1
    else:
        return n_jobs


def check_adjacency_vector(adjacency_vectors: Union[sparse.csr_matrix, np.ndarray],
                           n: Optional[int] = None) -> sparse.csr_matrix:
    """Check format of new samples for predict methods"""
    adjacency_vectors = check_format(adjacency_vectors)

    if n is not None:
        if adjacency_vectors.shape[1] != n:
            raise ValueError('The adjacency vector must be of length equal to the number nodes in the initial graph.')

    return adjacency_vectors


def check_n_clusters(n_clusters: int, n_row: int, n_min: int = 0):
    """Check that the number of clusters"""
    if n_clusters > n_row:
        raise ValueError('The number of clusters exceeds the number of rows.')
    if n_clusters < n_min:
        raise ValueError('The number of clusters must be at least {}.'.format(n_min))
    else:
        return


def check_min_size(n_row, n_min):
    """Check that an adjacency has the required number of rows and returns an error otherwise."""
    if n_row < n_min:
        raise ValueError('The graph must contain at least {} nodes.'.format(n_min))
    else:
        return


def check_dendrogram(dendrogram):
    """Check the shape of a dendrogram."""
    if dendrogram.ndim != 2 or dendrogram.shape[1] != 4:
        raise ValueError("Dendrogram has incorrect shape.")
    else:
        return


def check_min_nnz(nnz, n_min):
    """Check that an adjacency has the required number of edges and returns an error otherwise."""
    if nnz < n_min:
        raise ValueError('The graph must contain at least {} edge(s).'.format(n_min))
    else:
        return


def check_n_components(n_components, n_min) -> int:
    """Check the number of components"""
    if n_components > n_min:
        warnings.warn(Warning("The dimension of the embedding cannot exceed {}."
                              "Changed accordingly.".format(n_min)))
        return n_min
    else:
        return n_components


def check_scaling(scaling: float, adjacency: sparse.csr_matrix, regularize: bool):
    """Check the scaling factor"""
    if scaling < 0:
        raise ValueError("The 'scaling' parameter must be non-negative.")

    if scaling and (not regularize) and not is_connected(adjacency):
        raise ValueError("Positive 'scaling' is valid only if the graph is connected or with regularization."
                         "Call 'fit' either with 'scaling' = 0 or positive 'regularization'.")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""tests for force atlas2 embeddings"""
import unittest

import numpy as np

from sknetwork.data.test_graphs import test_graph, test_digraph
from sknetwork.embedding.force_atlas import ForceAtlas


class TestEmbeddings(unittest.TestCase):

    def test_options(self):
        for adjacency in [test_graph(), test_digraph()]:
            n = adjacency.shape[0]

            force_atlas = ForceAtlas()
            layout = force_atlas.fit_transform(adjacency)
            self.assertEqual((n, 2), layout.shape)

            force_atlas = ForceAtlas(lin_log=True)
            layout = force_atlas.fit_transform(adjacency)
            self.assertEqual((n, 2), layout.shape)

            force_atlas = ForceAtlas(approx_radius=1.)
            layout = force_atlas.fit_transform(adjacency)
            self.assertEqual((n, 2), layout.shape)

            force_atlas.fit(adjacency, pos_init=layout, n_iter=1)

    def test_errors(self):
        adjacency = test_graph()
        with self.assertRaises(ValueError):
            ForceAtlas().fit(adjacency, pos_init=np.ones((5, 7)))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on May 2020
@author: Nathan de Lara <ndelara@enst.fr>
"""
from typing import Union

import numpy as np
from scipy import sparse
from scipy.sparse.linalg import LinearOperator

from sknetwork.linalg.polynome import Polynome
from sknetwork.ranking.base import BaseRanking, BaseBiRanking
from sknetwork.utils.check import check_format
from sknetwork.utils.format import bipartite2undirected


class Katz(BaseRanking):
    """Katz centrality:

    :math:`\\sum_{k=1}^K\\alpha^k(A^k)^T\\mathbf{1}`.

    * Graphs
    * Digraphs

    Parameters
    ----------
    damping_factor : float
        Decay parameter for path contributions.
    path_length : int
        Maximum length of the paths to take into account.

    Examples
    --------
    >>> from sknetwork.data.toy_graphs import house
    >>> adjacency = house()
    >>> katz = Katz()
    >>> scores = katz.fit_transform(adjacency)
    >>> np.round(scores, 2)
    array([6.5 , 8.25, 5.62, 5.62, 8.25])

    References
    ----------
    Katz, L. (1953). `A new status index derived from sociometric analysis
    <https://link.springer.com/content/pdf/10.1007/BF02289026.pdf>`_. Psychometrika, 18(1), 39-43.
    """
    def __init__(self, damping_factor: float = 0.5, path_length: int = 4):
        super(Katz, self).__init__()
        self.damping_factor = damping_factor
        self.path_length = path_length

    def fit(self, adjacency: Union[sparse.csr_matrix, np.ndarray, LinearOperator]) -> 'Katz':
        """Katz centrality.

        Parameters
        ----------
        adjacency :
            Adjacency matrix of the graph.

        Returns
        -------
        self: :class:`Katz`
        """
        if not isinstance(adjacency, LinearOperator):
            adjacency = check_format(adjacency)
        n = adjacency.shape[0]
        coeffs = self.damping_factor ** np.arange(self.path_length + 1)
        coeffs[0] = 0.
        polynome = Polynome(adjacency.T.astype(bool).tocsr(), coeffs)

        self.scores_ = polynome.dot(np.ones(n))
        return self


class BiKatz(Katz, BaseBiRanking):
    """Katz centrality for bipartite graphs.

    * Bigraphs

    Parameters
    ----------
    damping_factor : float
        Decay parameter for path contributions.
    path_length : int
        Maximum length of the paths to take into account.

    Examples
    --------
    >>> from sknetwork.data.toy_graphs import star_wars
    >>> biadjacency = star_wars()
    >>> bikatz = BiKatz()
    >>> scores = bikatz.fit_transform(biadjacency)
    >>> np.round(scores, 2)
    array([6.38, 3.06, 8.81, 5.75])

    References
    ----------
    Katz, L. (1953). `A new status index derived from sociometric analysis
    <https://link.springer.com/content/pdf/10.1007/BF02289026.pdf>`_. Psychometrika, 18(1), 39-43.
    """
    def __init__(self, damping_factor: float = 0.5, path_length: int = 4):
        super(BiKatz, self).__init__(damping_factor=damping_factor, path_length=path_length)

    def fit(self, biadjacency: Union[sparse.csr_matrix, np.ndarray]) -> 'BiKatz':
        """Katz centrality.

        Parameters
        ----------
        biadjacency :
            Biadjacency matrix of the graph.

        Returns
        -------
        self: :class:`BiKatz`
        """
        biadjacency = check_format(biadjacency)
        n_row, n_col = biadjacency.shape
        adjacency = bipartite2undirected(biadjacency)

        Katz.fit(self, adjacency)
        self._split_vars(n_row)
        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on March 2019
@author: Thomas Bonald <bonald@enst.fr>
"""

import unittest

from sknetwork.data.test_graphs import *
from sknetwork.hierarchy import Paris, LouvainHierarchy, dasgupta_cost, dasgupta_score, tree_sampling_divergence


# noinspection PyMissingOrEmptyDocstring
class TestMetrics(unittest.TestCase):

    def setUp(self):
        self.paris = Paris()
        self.louvain_hierarchy = LouvainHierarchy()

    def test_undirected(self):
        adjacency = test_graph()
        dendrogram = self.paris.fit_transform(adjacency)
        self.assertAlmostEqual(dasgupta_cost(adjacency, dendrogram), 3.25, 2)
        self.assertAlmostEqual(dasgupta_score(adjacency, dendrogram), 0.675, 2)
        self.assertAlmostEqual(tree_sampling_divergence(adjacency, dendrogram), 0.533, 2)
        dendrogram = self.louvain_hierarchy.fit_transform(adjacency)
        self.assertAlmostEqual(dasgupta_cost(adjacency, dendrogram), 4.08, 2)
        self.assertAlmostEqual(dasgupta_score(adjacency, dendrogram), 0.592, 2)
        self.assertAlmostEqual(tree_sampling_divergence(adjacency, dendrogram), 0.485, 2)

    def test_directed(self):
        adjacency = test_digraph()
        dendrogram = self.paris.fit_transform(adjacency)
        self.assertAlmostEqual(dasgupta_score(adjacency, dendrogram), 0.672, 2)
        self.assertAlmostEqual(tree_sampling_divergence(adjacency, dendrogram), 0.484, 2)
        dendrogram = self.louvain_hierarchy.fit_transform(adjacency)
        self.assertAlmostEqual(dasgupta_score(adjacency, dendrogram), 0.627, 2)
        self.assertAlmostEqual(tree_sampling_divergence(adjacency, dendrogram), 0.463, 2)

    def test_disconnected(self):
        adjacency = test_graph_disconnect()
        dendrogram = self.paris.fit_transform(adjacency)
        self.assertAlmostEqual(dasgupta_score(adjacency, dendrogram), 0.752, 2)
        self.assertAlmostEqual(tree_sampling_divergence(adjacency, dendrogram), 0.627, 2)
        dendrogram = self.louvain_hierarchy.fit_transform(adjacency)
        self.assertAlmostEqual(dasgupta_score(adjacency, dendrogram), 0.691, 2)
        self.assertAlmostEqual(tree_sampling_divergence(adjacency, dendrogram), 0.549, 2)

    def test_options(self):
        adjacency = test_graph()
        dendrogram = self.paris.fit_transform(adjacency)
        self.assertAlmostEqual(dasgupta_score(adjacency, dendrogram, weights='degree'), 0.659, 2)
        self.assertAlmostEqual(tree_sampling_divergence(adjacency, dendrogram, weights='uniform'), 0.418, 2)
        self.assertAlmostEqual(tree_sampling_divergence(adjacency, dendrogram, normalized=False), 0.738, 2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Apr, 2019
@author: Nathan de Lara <ndelara@enst.fr>
"""
from typing import Optional, Union

import numpy as np

from sknetwork.utils.check import check_seeds


def stack_seeds(n_row: int, n_col: int, seeds_row: Optional[Union[np.ndarray, dict]],
                seeds_col: Optional[Union[np.ndarray, dict]] = None, default_value: float = -1) -> np.ndarray:
    """Process seeds for rows and columns and stack the results into a single vector."""
    if seeds_row is None and seeds_col is None:
        seeds_row = np.ones(n_row)
        seeds_col = default_value * np.ones(n_col)
    elif seeds_row is None:
        seeds_row = default_value * np.ones(n_row)
    elif seeds_col is None:
        seeds_col = default_value * np.ones(n_col)
    seeds_row = check_seeds(seeds_row, n_row)
    seeds_col = check_seeds(seeds_col, n_col)
    return np.hstack((seeds_row, seeds_col))


def seeds2probs(n: int, seeds: Union[dict, np.ndarray] = None) -> np.ndarray:
    """Transform seeds into probability vector.

    Parameters
    ----------
    n : int
        Total number of samples.
    seeds :
        If ``None``, the uniform distribution is used.
        Otherwise, a non-negative, non-zero vector or a dictionary must be provided.

    Returns
    -------
    probs: np.ndarray
        A probability vector.
    """
    if seeds is None:
        return np.ones(n) / n
    else:
        seeds = check_seeds(seeds, n)
        probs = np.zeros_like(seeds, dtype=float)
        ix = (seeds > 0)
        probs[ix] = seeds[ix]
        w: float = probs.sum()
        if w > 0:
            return probs / w
        else:
            raise ValueError('At least one seeds must have a positive probability.')



# -*- coding: utf-8 -*-
# tests for metrics.py
""""tests for clustering metrics"""
import unittest

import numpy as np

from sknetwork.clustering import modularity, bimodularity, comodularity, normalized_std, Louvain
from sknetwork.data import star_wars, karate_club
from sknetwork.data.test_graphs import test_graph


class TestClusteringMetrics(unittest.TestCase):

    def setUp(self):
        """Basic graph for tests"""
        self.adjacency = test_graph()
        n = self.adjacency.shape[0]
        labels = np.zeros(n)
        labels[0] = 1
        self.labels = labels.astype(int)
        self.unique_cluster = np.zeros(n, dtype=int)

    def test_api(self):
        for metric in [modularity, comodularity]:
            _, fit, div = metric(self.adjacency, self.labels, return_all=True)
            mod = metric(self.adjacency, self.labels, return_all=False)
            self.assertAlmostEqual(fit - div, mod)
            self.assertAlmostEqual(metric(self.adjacency, self.unique_cluster), 0.)

            with self.assertRaises(ValueError):
                metric(self.adjacency, self.labels[:3])

    def test_modularity(self):
        adjacency = karate_club()
        labels = Louvain().fit_transform(adjacency)
        self.assertAlmostEqual(0.42, modularity(adjacency, labels), places=2)

    def test_bimodularity(self):
        biadjacency = star_wars()
        labels_row = np.array([0, 0, 1, 1])
        labels_col = np.array([0, 1, 0])
        self.assertEqual(bimodularity(biadjacency, labels_row, labels_col), 0.1875)

        with self.assertRaises(ValueError):
            bimodularity(biadjacency, labels_row[:2], labels_col)
        with self.assertRaises(ValueError):
            bimodularity(biadjacency, labels_row, labels_col[:2])

    def test_nsd(self):
        balanced = np.arange(5)
        self.assertEqual(1, normalized_std(balanced))
        unbalanced = np.zeros(5)
        unbalanced[0] = 1
        self.assertGreaterEqual(normalized_std(unbalanced), 0)

        with self.assertRaises(ValueError):
            normalized_std(np.zeros(5))



# -*- coding: utf-8 -*-
# tests for test_graphs.py
"""
@author: Quentin Lutz <qlutz@enst.fr>
@author: Nathan de Lara <ndelara@enst.fr>
@author: Thomas Bonald <tbonald@enst.fr>
"""

import unittest

from sknetwork.data.test_graphs import *


class TestTestGraphs(unittest.TestCase):

    def test_undirected(self):
        adjacency = test_graph()
        self.assertEqual(adjacency.shape, (10, 10))
        adjacency = test_graph_disconnect()
        self.assertEqual(adjacency.shape, (10, 10))

    def test_directed(self):
        adjacency = test_digraph()
        self.assertEqual(adjacency.shape, (10, 10))

    def test_bipartite(self):
        biadjacency = test_bigraph()
        self.assertEqual(biadjacency.shape, (6, 8))
        biadjacency = test_bigraph_disconnect()
        self.assertEqual(biadjacency.shape, (6, 8))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""tests for embeddings"""

import unittest

from sknetwork.data.test_graphs import *
from sknetwork.embedding import Spectral, BiSpectral, SVD, GSVD, Spring


class TestEmbeddings(unittest.TestCase):

    def setUp(self):
        """Algorithms by input types."""
        self.methods = [Spectral(), GSVD(), SVD()]
        self.bimethods = [BiSpectral(), GSVD(), SVD()]

    def test_undirected(self):
        adjacency = test_graph()
        n = adjacency.shape[0]

        for method in self.methods:
            with self.assertRaises(ValueError):
                method.predict(adjacency[0])

            embedding = method.fit_transform(adjacency)

            self.assertEqual(embedding.shape, (n, 2))

            ref = embedding[0]
            pred1 = method.predict(adjacency[0])
            pred2 = method.predict(adjacency[0].toarray())

            self.assertEqual(pred1.shape, (2,))
            self.assertAlmostEqual(np.linalg.norm(pred1 - pred2), 0)
            self.assertAlmostEqual(np.linalg.norm(pred1 - ref), 0)

            pred1 = method.predict(adjacency)
            pred2 = method.predict(adjacency.toarray())

            self.assertTupleEqual(pred1.shape, (n, 2))
            self.assertAlmostEqual(np.linalg.norm(pred1 - pred2), 0)
            self.assertAlmostEqual(np.linalg.norm(pred1 - embedding), 0)

        method = Spring()
        embedding = method.fit_transform(adjacency)
        self.assertEqual(embedding.shape, (n, 2))
        pred1 = method.predict(adjacency[0])
        pred2 = method.predict(adjacency[0].toarray())
        self.assertEqual(pred1.shape, (2,))
        self.assertAlmostEqual(np.linalg.norm(pred1 - pred2), 0)

        pred1 = method.predict(adjacency)
        pred2 = method.predict(adjacency.toarray())
        self.assertTupleEqual(pred1.shape, (n, 2))
        self.assertAlmostEqual(np.linalg.norm(pred1 - pred2), 0)

    def test_bimethods(self):

        for adjacency in [test_digraph(), test_bigraph()]:
            n_row, n_col = adjacency.shape

            for method in self.bimethods:
                method.fit(adjacency)

                self.assertEqual(method.embedding_.shape, (n_row, 2))
                self.assertEqual(method.embedding_row_.shape, (n_row, 2))
                self.assertEqual(method.embedding_col_.shape, (n_col, 2))

                ref = method.embedding_[0]
                pred1 = method.predict(adjacency[0])
                pred2 = method.predict(adjacency[0].toarray())

                self.assertEqual(pred1.shape, (2,))
                self.assertAlmostEqual(np.linalg.norm(pred1 - pred2), 0)
                self.assertAlmostEqual(np.linalg.norm(pred1 - ref), 0)

                ref = method.embedding_
                pred1 = method.predict(adjacency)
                pred2 = method.predict(adjacency.toarray())

                self.assertTupleEqual(pred1.shape, (n_row, 2))
                self.assertAlmostEqual(np.linalg.norm(pred1 - pred2), 0)
                self.assertAlmostEqual(np.linalg.norm(pred1 - ref), 0)

    def test_disconnected(self):
        n = 10
        adjacency = np.eye(n)
        for method in self.methods:
            embedding = method.fit_transform(adjacency)
            self.assertEqual(embedding.shape, (n, 2))



"""clustering module"""
from sknetwork.clustering.base import BaseClustering, BaseBiClustering
from sknetwork.clustering.kmeans import BiKMeans, KMeans
from sknetwork.clustering.louvain import BiLouvain, Louvain
from sknetwork.clustering.metrics import modularity, bimodularity, comodularity, normalized_std
from sknetwork.clustering.postprocess import reindex_labels
from sknetwork.clustering.propagation_clustering import BiPropagationClustering, PropagationClustering



#!/usr/bin/env python3
# coding: utf-8
"""
Created on Thu Sep 13 2018
@author: Nathan de Lara <ndelara@enst.fr>
@author: Thomas Bonald <bonald@enst.fr>
"""
from typing import Union

import numpy as np
from scipy import sparse

from sknetwork.embedding.base import BaseEmbedding, BaseBiEmbedding
from sknetwork.linalg import EigSolver, HalkoEig, LanczosEig, auto_solver, diag_pinv, normalize, LaplacianOperator, \
    NormalizedAdjacencyOperator, RegularizedAdjacency
from sknetwork.utils.check import check_format, check_square, check_symmetry, check_adjacency_vector, \
    check_nonnegative, check_n_components, check_scaling
from sknetwork.utils.format import bipartite2undirected


def set_solver(solver: str, adjacency):
    """Eigenvalue solver based on keyword"""
    if solver == 'auto':
        solver: str = auto_solver(adjacency.nnz)
    if solver == 'lanczos':
        solver: EigSolver = LanczosEig()
    else:  # pragma: no cover
        solver: EigSolver = HalkoEig()
    return solver


class LaplacianEmbedding(BaseEmbedding):
    """Spectral embedding of graphs, based the spectral decomposition of the Laplacian matrix :math:`L = D - A`.
    Eigenvectors are considered in increasing order of eigenvalues, skipping the first eigenvector.

    * Graphs

    Parameters
    ----------
    n_components : int (default = 2)
        Dimension of the embedding space.
    regularization : ``None`` or float (default = ``0.01``)
        Add edges of given weight between all pairs of nodes.
    relative_regularization : bool (default = ``True``)
        If ``True``, consider the regularization as relative to the total weight of the graph.
    scaling : float (non-negative, default = ``0.5``)
        Scaling factor :math:`\\alpha` so that each component is divided by
        :math:`\\lambda^\\alpha`, with :math:`\\lambda` the corresponding eigenvalue of
        the Laplacian matrix :math:`L`. Require regularization if positive and the graph is not connected.
        The default value :math:`\\alpha=\\frac 1 2` equalizes the energy levels of
        the corresponding mechanical system.
    normalized : bool (default = ``True``)
        If ``True``, normalize the embedding so that each vector has norm 1 in the embedding space, i.e.,
        each vector lies on the unit sphere.
    solver : ``'auto'``, ``'halko'``, ``'lanczos'`` (default = ``'auto'``)
        Which eigenvalue solver to use.

        * ``'auto'`` call the auto_solver function.
        * ``'halko'``: randomized method, fast but less accurate than ``'lanczos'`` for ill-conditioned matrices.
        * ``'lanczos'``: power-iteration based method.

    Attributes
    ----------
    embedding_ : array, shape = (n, n_components)
        Embedding of the nodes.
    eigenvalues_ : array, shape = (n_components)
        Eigenvalues in increasing order (first eigenvalue ignored).
    eigenvectors_ : array, shape = (n, n_components)
        Corresponding eigenvectors.
    regularization_ : ``None`` or float
        Regularization factor added to all pairs of nodes.

    Example
    -------
    >>> from sknetwork.embedding import Spectral
    >>> from sknetwork.data import karate_club
    >>> laplacian = LaplacianEmbedding()
    >>> adjacency = karate_club()
    >>> embedding = laplacian.fit_transform(adjacency)
    >>> embedding.shape
    (34, 2)

    References
    ----------
    Belkin, M. & Niyogi, P. (2003). Laplacian Eigenmaps for Dimensionality Reduction and Data Representation,
    Neural computation.
    """
    def __init__(self, n_components: int = 2, regularization: Union[None, float] = 0.01,
                 relative_regularization: bool = True, scaling: float = 0.5,
                 normalized: bool = True, solver: str = 'auto'):
        super(LaplacianEmbedding, self).__init__()

        self.n_components = n_components
        self.regularization = None if regularization == 0 else regularization
        self.relative_regularization = relative_regularization
        self.scaling = scaling
        self.normalized = normalized
        self.solver = solver

        self.eigenvalues_ = None
        self.eigenvectors_ = None
        self.regularization_ = None

    def fit(self, adjacency: Union[sparse.csr_matrix, np.ndarray]) -> 'LaplacianEmbedding':
        """Compute the graph embedding.

        Parameters
        ----------
        adjacency :
              Adjacency matrix of the graph (symmetric matrix).

        Returns
        -------
        self: :class:`LaplacianEmbedding`
        """
        adjacency = check_format(adjacency).asfptype()
        check_square(adjacency)
        check_symmetry(adjacency)
        n = adjacency.shape[0]

        regularize: bool = not (self.regularization is None or self.regularization == 0.)
        check_scaling(self.scaling, adjacency, regularize)

        if regularize:
            solver: EigSolver = LanczosEig()
        else:
            solver = set_solver(self.solver, adjacency)
        n_components = 1 + check_n_components(self.n_components, n-2)

        weights = adjacency.dot(np.ones(n))
        regularization = self.regularization
        if regularization:
            if self.relative_regularization:
                regularization = regularization * weights.sum() / n ** 2
            weights += regularization * n
            laplacian = LaplacianOperator(adjacency, regularization)
        else:
            weight_diag = sparse.diags(weights, format='csr')
            laplacian = weight_diag - adjacency

        solver.which = 'SM'
        solver.fit(matrix=laplacian, n_components=n_components)
        eigenvalues = solver.eigenvalues_[1:]
        eigenvectors = solver.eigenvectors_[:, 1:]

        embedding = eigenvectors.copy()

        if self.scaling:
            eigenvalues_inv_diag = diag_pinv(eigenvalues ** self.scaling)
            embedding = eigenvalues_inv_diag.dot(embedding.T).T

        if self.normalized:
            embedding = normalize(embedding, p=2)

        self.embedding_ = embedding
        self.eigenvalues_ = eigenvalues
        self.eigenvectors_ = eigenvectors
        self.regularization_ = regularization

        return self


class Spectral(BaseEmbedding):
    """Spectral embedding of graphs, based the spectral decomposition of the transition matrix
    :math:`P = D^{-1}A`.
    Eigenvectors are considered in decreasing order of eigenvalues, skipping the first eigenvector.

    * Graphs

    See :class:`BiSpectral` for digraphs and bigraphs.

    Parameters
    ----------
    n_components : int (default = ``2``)
        Dimension of the embedding space.
    regularization : ``None`` or float (default = ``0.01``)
        Add edges of given weight between all pairs of nodes.
    relative_regularization : bool (default = ``True``)
        If ``True``, consider the regularization as relative to the total weight of the graph.
    scaling : float (non-negative, default = ``0.5``)
        Scaling factor :math:`\\alpha` so that each component is divided by
        :math:`(1 - \\lambda)^\\alpha`, with :math:`\\lambda` the corresponding eigenvalue of
        the transition matrix :math:`P`. Require regularization if positive and the graph is not connected.
        The default value :math:`\\alpha=\\frac 1 2` equalizes the energy levels of
        the corresponding mechanical system.
    normalized : bool (default = ``True``)
        If ``True``, normalize the embedding so that each vector has norm 1 in the embedding space, i.e.,
        each vector lies on the unit sphere.
    solver : ``'auto'``, ``'halko'``, ``'lanczos'`` (default = ``'auto'``)
        Which eigenvalue solver to use.

        * ``'auto'`` call the auto_solver function.
        * ``'halko'``: randomized method, fast but less accurate than ``'lanczos'`` for ill-conditioned matrices.
        * ``'lanczos'``: power-iteration based method.

    Attributes
    ----------
    embedding_ : array, shape = (n, n_components)
        Embedding of the nodes.
    eigenvalues_ : array, shape = (n_components)
        Eigenvalues in decreasing order (first eigenvalue ignored).
    eigenvectors_ : array, shape = (n, n_components)
        Corresponding eigenvectors.
    regularization_ : ``None`` or float
        Regularization factor added to all pairs of nodes.

    Example
    -------
    >>> from sknetwork.embedding import Spectral
    >>> from sknetwork.data import karate_club
    >>> spectral = Spectral()
    >>> adjacency = karate_club()
    >>> embedding = spectral.fit_transform(adjacency)
    >>> embedding.shape
    (34, 2)

    References
    ----------
    Belkin, M. & Niyogi, P. (2003). Laplacian Eigenmaps for Dimensionality Reduction and Data Representation,
    Neural computation.
    """
    def __init__(self, n_components: int = 2, regularization: Union[None, float] = 0.01,
                 relative_regularization: bool = True, scaling: float = 0.5,
                 normalized: bool = True, solver: str = 'auto'):
        super(Spectral, self).__init__()

        self.n_components = n_components
        self.regularization = None if regularization == 0 else regularization
        self.relative_regularization = relative_regularization
        self.scaling = scaling
        self.normalized = normalized
        self.solver = solver

        self.eigenvalues_ = None
        self.eigenvectors_ = None
        self.regularization_ = None

    def fit(self, adjacency: Union[sparse.csr_matrix, np.ndarray]) -> 'Spectral':
        """Compute the graph embedding.

        Parameters
        ----------
        adjacency :
              Adjacency matrix of the graph (symmetric matrix).

        Returns
        -------
        self: :class:`Spectral`
        """
        adjacency = check_format(adjacency).asfptype()
        check_square(adjacency)
        check_symmetry(adjacency)
        n = adjacency.shape[0]

        solver = set_solver(self.solver, adjacency)
        n_components = 1 + check_n_components(self.n_components, n-2)

        regularize: bool = not (self.regularization is None or self.regularization == 0.)
        check_scaling(self.scaling, adjacency, regularize)

        weights = adjacency.dot(np.ones(n))
        regularization = self.regularization
        if regularization:
            if self.relative_regularization:
                regularization = regularization * weights.sum() / n ** 2
            weights += regularization * n

        # Spectral decomposition of the normalized adjacency matrix
        weights_inv_sqrt_diag = diag_pinv(np.sqrt(weights))

        if regularization:
            norm_adjacency = NormalizedAdjacencyOperator(adjacency, regularization)
        else:
            norm_adjacency = weights_inv_sqrt_diag.dot(adjacency.dot(weights_inv_sqrt_diag))

        solver.which = 'LA'
        solver.fit(matrix=norm_adjacency, n_components=n_components)
        eigenvalues = solver.eigenvalues_
        index = np.argsort(-eigenvalues)[1:]  # skip first eigenvalue
        eigenvalues = eigenvalues[index]
        eigenvectors = weights_inv_sqrt_diag.dot(solver.eigenvectors_[:, index])

        embedding = eigenvectors.copy()

        if self.scaling:
            eigenvalues_inv_diag = diag_pinv((1 - eigenvalues) ** self.scaling)
            embedding = eigenvalues_inv_diag.dot(embedding.T).T

        if self.normalized:
            embedding = normalize(embedding, p=2)

        self.embedding_ = embedding
        self.eigenvalues_ = eigenvalues
        self.eigenvectors_ = eigenvectors
        self.regularization_ = regularization

        return self

    def predict(self, adjacency_vectors: Union[sparse.csr_matrix, np.ndarray]) -> np.ndarray:
        """Predict the embedding of new nodes, defined by their adjacency vectors.

        Parameters
        ----------
        adjacency_vectors :
            Adjacency vectors of nodes.
            Array of shape (n_col,) (single vector) or (n_vectors, n_col)

        Returns
        -------
        embedding_vectors : np.ndarray
            Embedding of the nodes.
        """
        self._check_fitted()
        eigenvectors = self.eigenvectors_
        eigenvalues = self.eigenvalues_
        n = eigenvectors.shape[0]

        adjacency_vectors = check_adjacency_vector(adjacency_vectors, n)
        check_nonnegative(adjacency_vectors)

        # regularization
        if self.regularization_:
            adjacency_vectors = RegularizedAdjacency(adjacency_vectors, self.regularization_)

        # projection in the embedding space
        averaging = normalize(adjacency_vectors, p=1)
        embedding_vectors = averaging.dot(eigenvectors)
        embedding_vectors = diag_pinv(eigenvalues).dot(embedding_vectors.T).T

        if self.scaling:
            eigenvalues_inv_diag = diag_pinv((1 - eigenvalues) ** self.scaling)
            embedding_vectors = eigenvalues_inv_diag.dot(embedding_vectors.T).T

        if self.normalized:
            embedding_vectors = normalize(embedding_vectors, p=2)

        if embedding_vectors.shape[0] == 1:
            embedding_vectors = embedding_vectors.ravel()

        return embedding_vectors


class BiSpectral(Spectral, BaseBiEmbedding):
    """Spectral embedding of bipartite and directed graphs, based the spectral embedding
    of the corresponding undirected graph, with adjacency matrix:

        :math:`A  = \\begin{bmatrix} 0 & B \\\\ B^T & 0 \\end{bmatrix}`

    where :math:`B` is the biadjacency matrix of the bipartite graph or the adjacency matrix
    of the directed graph.

    * Digraphs
    * Bigraphs

    Parameters
    ----------
    n_components : int (default = 2)
        Dimension of the embedding space.
    regularization : ``None`` or float (default = ``0.01``)
        Add edges of given weight between all pairs of nodes.
    relative_regularization : bool (default = ``True``)
        If ``True``, consider the regularization as relative to the total weight of the graph.
    scaling : float (non-negative, default = ``0.5``)
        Scaling factor :math:`\\alpha` so that each component is divided by
        :math:`(1 - \\lambda)^\\alpha`, with :math:`\\lambda` the corresponding eigenvalue of
        the transition matrix :math:`P`. Require regularization if positive and the graph is not connected.
        The default value :math:`\\alpha=\\frac 1 2` equalizes the energy levels of
        the corresponding mechanical system.
    normalized : bool (default = ``True``)
        If ``True``, normalized the embedding so that each vector has norm 1 in the embedding space, i.e.,
        each vector lies on the unit sphere.
    solver : ``'auto'``, ``'halko'``, ``'lanczos'`` (default = ``'auto'``)
        Which eigenvalue solver to use.

        * ``'auto'`` call the auto_solver function.
        * ``'halko'``: randomized method, fast but less accurate than ``'lanczos'`` for ill-conditioned matrices.
        * ``'lanczos'``: power-iteration based method.

    Attributes
    ----------
    embedding_ : array, shape = (n_row, n_components)
        Embedding of the rows.
    embedding_row_ : array, shape = (n_row, n_components)
        Embedding of the rows (copy of **embedding_**).
    embedding_col_ : array, shape = (n_col, n_components)
        Embedding of the columns.
    eigenvalues_ : array, shape = (n_components)
        Eigenvalues in increasing order (first eigenvalue ignored).
    eigenvectors_ : array, shape = (n, n_components)
        Corresponding eigenvectors.
    regularization_ : ``None`` or float
        Regularization factor added to all pairs of nodes.

    Example
    -------
    >>> from sknetwork.embedding import BiSpectral
    >>> from sknetwork.data import movie_actor
    >>> bispectral = BiSpectral()
    >>> biadjacency = movie_actor()
    >>> embedding = bispectral.fit_transform(biadjacency)
    >>> embedding.shape
    (15, 2)

    References
    ----------
    Belkin, M. & Niyogi, P. (2003). Laplacian Eigenmaps for Dimensionality Reduction and Data Representation,
    Neural computation.
    """
    def __init__(self, n_components: int = 2, regularization: Union[None, float] = 0.01,
                 relative_regularization: bool = True, scaling: float = 0.5,
                 normalized: bool = True, solver: str = 'auto'):
        super(BiSpectral, self).__init__(n_components, regularization, relative_regularization, scaling,
                                         normalized, solver)

    def fit(self, biadjacency: Union[sparse.csr_matrix, np.ndarray]) -> 'BiSpectral':
        """Spectral embedding of the bipartite graph considered as undirected, with adjacency matrix:

        :math:`A  = \\begin{bmatrix} 0 & B \\\\ B^T & 0 \\end{bmatrix}`

        where :math:`B` is the biadjacency matrix (or the adjacency matrix of a directed graph).

        Parameters
        ----------
        biadjacency:
            Biadjacency matrix of the graph.

        Returns
        -------
        self: :class:`BiSpectral`
        """
        biadjacency = check_format(biadjacency)
        n_row, _ = biadjacency.shape
        Spectral.fit(self, bipartite2undirected(biadjacency))
        self._split_vars(n_row)

        return self

    def predict(self, adjacency_vectors: Union[sparse.csr_matrix, np.ndarray]) -> np.ndarray:
        """Predict the embedding of new rows, defined by their adjacency vectors.

        Parameters
        ----------
        adjacency_vectors :
            Adjacency vectors of nodes.
            Array of shape (n_col,) (single vector) or (n_vectors, n_col)

        Returns
        -------
        embedding_vectors : np.ndarray
            Embedding of the nodes.
        """
        self._check_fitted()
        n_row, _ = self.embedding_row_.shape
        n_col, _ = self.embedding_col_.shape

        adjacency_vectors = check_adjacency_vector(adjacency_vectors, n_col)
        empty_block = sparse.csr_matrix((adjacency_vectors.shape[0], n_row))
        adjacency_vectors = sparse.bmat([[empty_block, adjacency_vectors]], format='csr')
        embedding_vectors = Spectral.predict(self, adjacency_vectors)

        return embedding_vectors



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on November 12 2019
@author: Quentin Lutz <qlutz@enst.fr>
"""
from math import log
from typing import Union, Optional

import numpy as np
from scipy import sparse

from sknetwork.path.shortest_path import distance
from sknetwork.ranking.base import BaseRanking
from sknetwork.utils.check import check_format, check_square, check_connected


class Closeness(BaseRanking):
    """Closeness centrality of each node in a connected graph, corresponding to the average length of the
    shortest paths from that node to all the other ones.

    For a directed graph, the closeness centrality is computed in terms of outgoing paths.

    * Graphs
    * Digraphs

    Parameters
    ----------
    method :
        Denotes if the results should be exact or approximate.
    tol: float
        If ``method=='approximate'``, the allowed tolerance on each score entry.
    n_jobs:
        If an integer value is given, denotes the number of workers to use (-1 means the maximum number will be used).
        If ``None``, no parallel computations are made.

    Attributes
    ----------
    scores_ : np.ndarray
        Closeness centrality of each node.

    Example
    -------
    >>> from sknetwork.ranking import Closeness
    >>> from sknetwork.data import cyclic_digraph
    >>> closeness = Closeness()
    >>> adjacency = cyclic_digraph(3)
    >>> scores = closeness.fit_transform(adjacency)
    >>> np.round(scores, 2)
    array([0.67, 0.67, 0.67])

    References
    ----------
    Eppstein, D., & Wang, J. (2001, January).
    `Fast approximation of centrality.
    <http://jgaa.info/accepted/2004/EppsteinWang2004.8.1.pdf>`_
    In Proceedings of the twelfth annual ACM-SIAM symposium on Discrete algorithms (pp. 228-229).
    Society for Industrial and Applied Mathematics.
    """

    def __init__(self, method: str = 'exact', tol: float = 1e-1, n_jobs: Optional[int] = None):
        super(Closeness, self).__init__()

        self.method = method
        self.tol = tol
        self.n_jobs = n_jobs

    def fit(self, adjacency: Union[sparse.csr_matrix, np.ndarray]) -> 'Closeness':
        """Closeness centrality for connected graphs.

        Parameters
        ----------
        adjacency :
            Adjacency matrix of the graph.

        Returns
        -------
        self: :class:`Closeness`
        """
        adjacency = check_format(adjacency)
        check_square(adjacency)
        check_connected(adjacency)
        n = adjacency.shape[0]

        if self.method == 'exact':
            n_sources = n
            sources = np.arange(n)
        elif self.method == 'approximate':
            n_sources = min(int(log(n) / self.tol ** 2), n)
            sources = np.random.choice(np.arange(n), n_sources, replace=False)
        else:
            raise ValueError("Method should be either 'exact' or 'approximate'.")

        dists = distance(adjacency, n_jobs=self.n_jobs, sources=sources)

        self.scores_ = ((n - 1) * n_sources / n) / dists.T.dot(np.ones(n_sources))

        return self



"""Path module"""
from sknetwork.path.metrics import diameter
from sknetwork.path.search import breadth_first_search, depth_first_search
from sknetwork.path.shortest_path import distance, shortest_path



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Apr 2020
@author: Nathan de Lara <ndelara@enst.fr>
"""
import unittest

import numpy as np
from scipy import sparse

from sknetwork.data import movie_actor
from sknetwork.linalg import CoNeighborOperator, normalize


class TestOperators(unittest.TestCase):

    def test_coneighbors(self):
        biadjacency = movie_actor(metadata=False)
        operator = CoNeighborOperator(biadjacency)
        transition = normalize(operator)
        x = transition.dot(np.ones(transition.shape[1]))

        self.assertAlmostEqual(np.linalg.norm(x - np.ones(operator.shape[0])), 0)
        operator.astype(np.float)
        operator.right_sparse_dot(sparse.eye(operator.shape[1], format='csr'))

        operator1 = CoNeighborOperator(biadjacency, normalized=False)
        operator2 = CoNeighborOperator(biadjacency, normalized=False)
        x = np.random.randn(operator.shape[1])
        x1 = (-operator1).dot(x)
        x2 = (operator2 * -1).dot(x)
        x3 = operator1.T.dot(x)
        self.assertAlmostEqual(np.linalg.norm(x1 - x2), 0)
        self.assertAlmostEqual(np.linalg.norm(x2 - x3), 0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on October 2019
@author: Nathan de Lara <ndelara@enst.fr>
"""

import unittest

from sknetwork.data.test_graphs import test_graph
from sknetwork.embedding import Spectral
from sknetwork.hierarchy import Ward


class TestWard(unittest.TestCase):

    def test_options(self):
        adjacency = test_graph()
        ward = Ward(embedding_method=Spectral(3))
        dendrogram = ward.fit_transform(adjacency)
        self.assertEqual(dendrogram.shape, (adjacency.shape[0] - 1, 4))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on May, 2020
@author: Nathan de Lara <ndelara@enst.fr>
"""
from abc import ABC
from typing import Union, Iterable

import numpy as np
from scipy import sparse
from sknetwork.linkpred.first_order_core import common_neighbors_node_core, jaccard_node_core, salton_node_core, \
    sorensen_node_core, hub_promoted_node_core, hub_depressed_node_core, adamic_adar_node_core, \
    resource_allocation_node_core, \
    common_neighbors_edges_core, jaccard_edges_core, salton_edges_core, sorensen_edges_core, hub_promoted_edges_core, \
    hub_depressed_edges_core, adamic_adar_edges_core, resource_allocation_edges_core

from sknetwork.linkpred.base import BaseLinkPred
from sknetwork.utils.check import check_format


class FirstOrder(BaseLinkPred, ABC):
    """Base class for first order algorithms."""

    def __init__(self):
        super(FirstOrder, self).__init__()
        self.indptr_ = None
        self.indices_ = None

    def fit(self, adjacency: Union[sparse.csr_matrix, np.ndarray]):
        """Fit algorithm to the data.

        Parameters
        ----------
        adjacency :
            Adjacency matrix of the graph

        Returns
        -------
        self : :class:`FirstOrder`
        """
        adjacency = check_format(adjacency)
        adjacency.sort_indices()
        self.indptr_ = adjacency.indptr.astype(np.int32)
        self.indices_ = adjacency.indices.astype(np.int32)

        return self

    def _predict_node(self, source: int):
        """Prediction for a single node."""
        n = self.indptr_.shape[0] - 1
        return self._predict_base(source, np.arange(n))


class CommonNeighbors(FirstOrder):
    """Link prediction by common neighbors:

    :math:`s(i, j) = |\\Gamma_i \\cap \\Gamma_j|`.

    Attributes
    ----------
    indptr_ : np.ndarray
        Pointer index for neighbors.
    indices_ : np.ndarray
        Concatenation of neighbors.

    Examples
    --------
    >>> from sknetwork.data import house
    >>> adjacency = house()
    >>> cn = CommonNeighbors()
    >>> similarities = cn.fit_predict(adjacency, 0)
    >>> similarities
    array([2, 1, 1, 1, 1])
    >>> similarities = cn.predict([0, 1])
    >>> similarities
    array([[2, 1, 1, 1, 1],
           [1, 3, 0, 2, 1]])
    >>> similarities = cn.predict((0, 1))
    >>> similarities
    1
    >>> similarities = cn.predict([(0, 1), (1, 2)])
    >>> similarities
    array([1, 0])
    """
    def __init__(self):
        super(CommonNeighbors, self).__init__()

    def _predict_base(self, source: int, targets: Iterable):
        """Prediction for a single node."""
        return np.asarray(common_neighbors_node_core(self.indptr_, self.indices_, np.int32(source),
                                                     np.array(targets, dtype=np.int32))).astype(int)

    def _predict_edges(self, edges: np.ndarray):
        """Prediction for multiple edges."""
        return np.asarray(common_neighbors_edges_core(self.indptr_, self.indices_, edges.astype(np.int32))).astype(int)


class JaccardIndex(FirstOrder):
    """Link prediction by Jaccard Index:

    :math:`s(i, j) = \\dfrac{|\\Gamma_i \\cap \\Gamma_j|}{|\\Gamma_i \\cup \\Gamma_j|}`.

    Attributes
    ----------
    indptr_ : np.ndarray
        Pointer index for neighbors.
    indices_ : np.ndarray
        Concatenation of neighbors.

    Examples
    --------
    >>> from sknetwork.data import house
    >>> adjacency = house()
    >>> jaccard = JaccardIndex()
    >>> similarities = jaccard.fit_predict(adjacency, 0)
    >>> similarities.round(2)
    array([1.  , 0.25, 0.33, 0.33, 0.25])
    >>> similarities = jaccard.predict([0, 1])
    >>> similarities.round(2)
    array([[1.  , 0.25, 0.33, 0.33, 0.25],
           [0.25, 1.  , 0.  , 0.67, 0.2 ]])
    >>> similarities = jaccard.predict((0, 1))
    >>> similarities.round(2)
    0.25
    >>> similarities = jaccard.predict([(0, 1), (1, 2)])
    >>> similarities.round(2)
    array([0.25, 0.  ])

    References
    ----------
    Jaccard, P. (1901) étude comparative de la distribution florale dans une portion des Alpes et du Jura.
    Bulletin de la Société Vaudoise des Sciences Naturelles, 37, 547-579.
    """
    def __init__(self):
        super(JaccardIndex, self).__init__()

    def _predict_base(self, source: int, targets: Iterable):
        """Prediction for a single node."""
        return np.asarray(jaccard_node_core(self.indptr_, self.indices_, np.int32(source), np.array(targets, dtype=np.int32)))

    def _predict_edges(self, edges: np.ndarray):
        """Prediction for multiple edges."""
        return np.asarray(jaccard_edges_core(self.indptr_, self.indices_, edges.astype(np.int32)))


class SaltonIndex(FirstOrder):
    """Link prediction by Salton Index:

    :math:`s(i, j) = \\dfrac{|\\Gamma_i \\cap \\Gamma_j|}{\\sqrt{|\\Gamma_i|.|\\Gamma_j|}}`.

    Attributes
    ----------
    indptr_ : np.ndarray
        Pointer index for neighbors.
    indices_ : np.ndarray
        Concatenation of neighbors.

    Examples
    --------
    >>> from sknetwork.data import house
    >>> adjacency = house()
    >>> salton = SaltonIndex()
    >>> similarities = salton.fit_predict(adjacency, 0)
    >>> similarities.round(2)
    array([1.  , 0.41, 0.5 , 0.5 , 0.41])
    >>> similarities = salton.predict([0, 1])
    >>> similarities.round(2)
    array([[1.  , 0.41, 0.5 , 0.5 , 0.41],
           [0.41, 1.  , 0.  , 0.82, 0.33]])
    >>> similarities = salton.predict((0, 1))
    >>> similarities.round(2)
    0.41
    >>> similarities = salton.predict([(0, 1), (1, 2)])
    >>> similarities.round(2)
    array([0.41, 0.  ])

    References
    ----------
    Martínez, V., Berzal, F., & Cubero, J. C. (2016).
    `A survey of link prediction in complex networks.
    <https://dl.acm.org/doi/pdf/10.1145/3012704>`_
    ACM computing surveys (CSUR), 49(4), 1-33.
    """
    def __init__(self):
        super(SaltonIndex, self).__init__()

    def _predict_base(self, source: int, targets: Iterable):
        """Prediction for a single node."""
        return np.asarray(salton_node_core(self.indptr_, self.indices_, np.int32(source), np.array(targets, dtype=np.int32)))

    def _predict_edges(self, edges: np.ndarray):
        """Prediction for multiple edges."""
        return np.asarray(salton_edges_core(self.indptr_, self.indices_, edges.astype(np.int32)))


class SorensenIndex(FirstOrder):
    """Link prediction by Salton Index:

    :math:`s(i, j) = \\dfrac{2|\\Gamma_i \\cap \\Gamma_j|}{|\\Gamma_i|+|\\Gamma_j|}`.

    Attributes
    ----------
    indptr_ : np.ndarray
        Pointer index for neighbors.
    indices_ : np.ndarray
        Concatenation of neighbors.

    Examples
    --------
    >>> from sknetwork.data import house
    >>> adjacency = house()
    >>> sorensen = SorensenIndex()
    >>> similarities = sorensen.fit_predict(adjacency, 0)
    >>> similarities.round(2)
    array([1. , 0.4, 0.5, 0.5, 0.4])
    >>> similarities = sorensen.predict([0, 1])
    >>> similarities.round(2)
    array([[1.  , 0.4 , 0.5 , 0.5 , 0.4 ],
           [0.4 , 1.  , 0.  , 0.8 , 0.33]])
    >>> similarities = sorensen.predict((0, 1))
    >>> similarities.round(2)
    0.4
    >>> similarities = sorensen.predict([(0, 1), (1, 2)])
    >>> similarities.round(2)
    array([0.4, 0. ])

    References
    ----------

    * Sørensen, T. J. (1948). A method of establishing groups of equal amplitude in plant sociology
      based on similarity of species content and its application to analyses of the vegetation on Danish commons.
      I kommission hos E. Munksgaard.
    * Martínez, V., Berzal, F., & Cubero, J. C. (2016).
      `A survey of link prediction in complex networks.
      <https://dl.acm.org/doi/pdf/10.1145/3012704>`_
      ACM computing surveys (CSUR), 49(4), 1-33.

    """
    def __init__(self):
        super(SorensenIndex, self).__init__()

    def _predict_base(self, source: int, targets: Iterable):
        """Prediction for a single node."""
        return np.asarray(sorensen_node_core(self.indptr_, self.indices_, np.int32(source), np.array(targets, dtype=np.int32)))

    def _predict_edges(self, edges: np.ndarray):
        """Prediction for multiple edges."""
        return np.asarray(sorensen_edges_core(self.indptr_, self.indices_, edges.astype(np.int32)))


class HubPromotedIndex(FirstOrder):
    """Link prediction by Hub Promoted Index:

    :math:`s(i, j) = \\dfrac{2|\\Gamma_i \\cap \\Gamma_j|}{min(|\\Gamma_i|,|\\Gamma_j|)}`.

    Attributes
    ----------
    indptr_ : np.ndarray
        Pointer index for neighbors.
    indices_ : np.ndarray
        Concatenation of neighbors.

    Examples
    --------
    >>> from sknetwork.data import house
    >>> adjacency = house()
    >>> hpi = HubPromotedIndex()
    >>> similarities = hpi.fit_predict(adjacency, 0)
    >>> similarities.round(2)
    array([1. , 0.5, 0.5, 0.5, 0.5])
    >>> similarities = hpi.predict([0, 1])
    >>> similarities.round(2)
    array([[1.  , 0.5 , 0.5 , 0.5 , 0.5 ],
           [0.5 , 1.  , 0.  , 1.  , 0.33]])
    >>> similarities = hpi.predict((0, 1))
    >>> similarities.round(2)
    0.5
    >>> similarities = hpi.predict([(0, 1), (1, 2)])
    >>> similarities.round(2)
    array([0.5, 0. ])

    References
    ----------
    Ravasz, E., Somera, A. L., Mongru, D. A., Oltvai, Z. N., & Barabási, A. L. (2002).
    `Hierarchical organization of modularity in metabolic networks.
    <https://arxiv.org/pdf/cond-mat/0209244.pdf>`_
    science, 297(5586), 1551-1555.
    """
    def __init__(self):
        super(HubPromotedIndex, self).__init__()

    def _predict_base(self, source: int, targets: Iterable):
        """Prediction for a single node."""
        return np.asarray(hub_promoted_node_core(self.indptr_, self.indices_, np.int32(source),
                                                 np.array(targets, dtype=np.int32)))

    def _predict_edges(self, edges: np.ndarray):
        """Prediction for multiple edges."""
        return np.asarray(hub_promoted_edges_core(self.indptr_, self.indices_, edges.astype(np.int32)))


class HubDepressedIndex(FirstOrder):
    """Link prediction by Hub Depressed Index:

    :math:`s(i, j) = \\dfrac{2|\\Gamma_i \\cap \\Gamma_j|}{max(|\\Gamma_i|,|\\Gamma_j|)}`.

    Attributes
    ----------
    indptr_ : np.ndarray
        Pointer index for neighbors.
    indices_ : np.ndarray
        Concatenation of neighbors.

    Examples
    --------
    >>> from sknetwork.data import house
    >>> adjacency = house()
    >>> hdi = HubDepressedIndex()
    >>> similarities = hdi.fit_predict(adjacency, 0)
    >>> similarities.round(2)
    array([1.  , 0.33, 0.5 , 0.5 , 0.33])
    >>> similarities = hdi.predict([0, 1])
    >>> similarities.round(2)
    array([[1.  , 0.33, 0.5 , 0.5 , 0.33],
           [0.33, 1.  , 0.  , 0.67, 0.33]])
    >>> similarities = hdi.predict((0, 1))
    >>> similarities.round(2)
    0.33
    >>> similarities = hdi.predict([(0, 1), (1, 2)])
    >>> similarities.round(2)
    array([0.33, 0.  ])

    References
    ----------
    Ravasz, E., Somera, A. L., Mongru, D. A., Oltvai, Z. N., & Barabási, A. L. (2002).
    `Hierarchical organization of modularity in metabolic networks.
    <https://arxiv.org/pdf/cond-mat/0209244.pdf>`_
    science, 297(5586), 1551-1555.
    """
    def __init__(self):
        super(HubDepressedIndex, self).__init__()

    def _predict_base(self, source: int, targets: Iterable):
        """Prediction for a single node."""
        return np.asarray(hub_depressed_node_core(self.indptr_, self.indices_, np.int32(source),
                                                  np.array(targets, dtype=np.int32)))

    def _predict_edges(self, edges: np.ndarray):
        """Prediction for multiple edges."""
        return np.asarray(hub_depressed_edges_core(self.indptr_, self.indices_, edges.astype(np.int32)))


class AdamicAdar(FirstOrder):
    """Link prediction by Adamic-Adar index:

    :math:`s(i, j) = \\underset{z \\in \\Gamma_i \\cap \\Gamma_j}{\\sum} \\dfrac{1}{\\log |\\Gamma_z|}`.

    Attributes
    ----------
    indptr_ : np.ndarray
        Pointer index for neighbors.
    indices_ : np.ndarray
        Concatenation of neighbors.

    Examples
    --------
    >>> from sknetwork.data import house
    >>> adjacency = house()
    >>> aa = AdamicAdar()
    >>> similarities = aa.fit_predict(adjacency, 0)
    >>> similarities.round(2)
    array([1.82, 0.91, 0.91, 0.91, 0.91])
    >>> similarities = aa.predict([0, 1])
    >>> similarities.round(2)
    array([[1.82, 0.91, 0.91, 0.91, 0.91],
           [0.91, 3.8 , 0.  , 2.35, 1.44]])
    >>> similarities = aa.predict((0, 1))
    >>> similarities.round(2)
    0.91
    >>> similarities = aa.predict([(0, 1), (1, 2)])
    >>> similarities.round(2)
    array([0.91, 0.  ])

    References
    ----------
    Adamic, L. A., & Adar, E. (2003). `Friends and neighbors on the web.
    <https://www.sciencedirect.com/science/article/pii/S0378873303000091>`_
    Social networks, 25(3), 211-230.
    """
    def __init__(self):
        super(AdamicAdar, self).__init__()

    def _predict_base(self, source: int, targets: Iterable):
        """Prediction for a single node."""
        return np.asarray(adamic_adar_node_core(self.indptr_, self.indices_, np.int32(source),
                                                np.array(targets, dtype=np.int32)))

    def _predict_edges(self, edges: np.ndarray):
        """Prediction for multiple edges."""
        return np.asarray(adamic_adar_edges_core(self.indptr_, self.indices_, edges.astype(np.int32)))


class ResourceAllocation(FirstOrder):
    """Link prediction by Resource Allocation index:

    :math:`s(i, j) = \\underset{z \\in \\Gamma_i \\cap \\Gamma_j}{\\sum} \\dfrac{1}{|\\Gamma_z|}`.

    Attributes
    ----------
    indptr_ : np.ndarray
        Pointer index for neighbors.
    indices_ : np.ndarray
        Concatenation of neighbors.

    Examples
    --------
    >>> from sknetwork.data import house
    >>> adjacency = house()
    >>> ra = ResourceAllocation()
    >>> similarities = ra.fit_predict(adjacency, 0)
    >>> similarities.round(2)
    array([0.67, 0.33, 0.33, 0.33, 0.33])
    >>> similarities = ra.predict([0, 1])
    >>> similarities.round(2)
    array([[0.67, 0.33, 0.33, 0.33, 0.33],
           [0.33, 1.33, 0.  , 0.83, 0.5 ]])
    >>> similarities = ra.predict((0, 1))
    >>> similarities.round(2)
    0.33
    >>> similarities = ra.predict([(0, 1), (1, 2)])
    >>> similarities.round(2)
    array([0.33, 0.  ])

    References
    ----------
    Zhou, T., Lü, L., & Zhang, Y. C. (2009).
    `Predicting missing links via local information.
    <https://link.springer.com/content/pdf/10.1140/epjb/e2009-00335-8.pdf>`_
    The European Physical Journal B, 71(4), 623-630.
    """
    def __init__(self):
        super(ResourceAllocation, self).__init__()

    def _predict_base(self, source: int, targets: Iterable):
        """Prediction for a single node."""
        return np.asarray(resource_allocation_node_core(self.indptr_, self.indices_, np.int32(source),
                                                        np.array(targets, dtype=np.int32)))

    def _predict_edges(self, edges: np.ndarray):
        """Prediction for multiple edges."""
        return np.asarray(resource_allocation_edges_core(self.indptr_, self.indices_, edges.astype(np.int32)))


class PreferentialAttachment(FirstOrder):
    """Link prediction by Preferential Attachment index:

    :math:`s(i, j) = |\\Gamma_i||\\Gamma_j|`.

    Attributes
    ----------
    indptr_ : np.ndarray
        Pointer index for neighbors.
    indices_ : np.ndarray
        Concatenation of neighbors.

    Examples
    --------
    >>> from sknetwork.data import house
    >>> adjacency = house()
    >>> pa = PreferentialAttachment()
    >>> similarities = pa.fit_predict(adjacency, 0)
    >>> similarities
    array([4, 6, 4, 4, 6], dtype=int32)
    >>> similarities = pa.predict([0, 1])
    >>> similarities
    array([[4, 6, 4, 4, 6],
           [6, 9, 6, 6, 9]], dtype=int32)
    >>> similarities = pa.predict((0, 1))
    >>> similarities
    6
    >>> similarities = pa.predict([(0, 1), (1, 2)])
    >>> similarities
    array([6, 6], dtype=int32)

    References
    ----------
    Albert, R., Barabási, L. (2002). `Statistical mechanics of complex networks
    <https://journals.aps.org/rmp/abstract/10.1103/RevModPhys.74.47>`_
    Reviews of Modern Physics.
    """
    def __init__(self):
        super(PreferentialAttachment, self).__init__()

    def _predict_base(self, source: int, targets: Iterable):
        """Prediction for a single node."""
        deg_src = self.indptr_[source+1] - self.indptr_[source]
        tgt = np.array(targets)
        deg_tgt = self.indptr_[tgt+1] - self.indptr_[tgt]
        return deg_src * deg_tgt

    def _predict_edges(self, edges: np.ndarray):
        """Prediction for multiple edges."""
        deg_src = self.indptr_[edges[:, 0] + 1] - self.indptr_[edges[:, 0]]
        deg_tgt = self.indptr_[edges[:, 1] + 1] - self.indptr_[edges[:, 1]]
        return deg_src * deg_tgt



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""tests for svd"""

import unittest

import numpy as np

from sknetwork.data import movie_actor
from sknetwork.linalg import LanczosSVD, HalkoSVD, SparseLR


def svd_err(matrix, u, v, sigma):
    """Approximation error for singular vectors."""
    err = matrix.dot(v) - u * sigma
    return np.linalg.norm(err)


# noinspection DuplicatedCode
class TestSolvers(unittest.TestCase):

    def setUp(self):
        """Simple biadjacency for tests."""
        self.biadjacency = movie_actor()
        n_row, n_col = self.biadjacency.shape
        self.slr = SparseLR(self.biadjacency, [(np.random.rand(n_row), np.random.rand(n_col))])

    def test_lanczos(self):
        solver = LanczosSVD()
        solver.fit(self.biadjacency, 2)
        self.assertEqual(len(solver.singular_values_), 2)
        self.assertAlmostEqual(svd_err(self.biadjacency, solver.singular_vectors_left_, solver.singular_vectors_right_,
                                       solver.singular_values_), 0)

        solver.fit(self.slr, 2)
        self.assertEqual(len(solver.singular_values_), 2)
        self.assertAlmostEqual(svd_err(self.slr, solver.singular_vectors_left_, solver.singular_vectors_right_,
                                       solver.singular_values_), 0)

    def test_halko(self):
        solver = HalkoSVD()
        solver.fit(self.biadjacency, 2)
        self.assertEqual(len(solver.singular_values_), 2)
        self.assertAlmostEqual(svd_err(self.biadjacency, solver.singular_vectors_left_, solver.singular_vectors_right_,
                                       solver.singular_values_), 0)

        solver.fit(self.slr, 2)
        self.assertEqual(len(solver.singular_values_), 2)
        self.assertAlmostEqual(svd_err(self.slr, solver.singular_vectors_left_, solver.singular_vectors_right_,
                                       solver.singular_values_), 0)

    def test_compare_solvers(self):
        lanczos = LanczosSVD()
        halko = HalkoSVD()

        lanczos.fit(self.biadjacency, 2)
        halko.fit(self.biadjacency, 2)
        self.assertAlmostEqual(np.linalg.norm(lanczos.singular_values_ - halko.singular_values_), 0.)

        lanczos.fit(self.slr, 2)
        halko.fit(self.slr, 2)
        self.assertAlmostEqual(np.linalg.norm(lanczos.singular_values_ - halko.singular_values_), 0.)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Jun 3, 2020
@author: Nathan de Lara <ndelara@enst.fr>
"""
import numpy as np
from scipy import sparse
from sknetwork.topology.dag_core import fit_core

from sknetwork.utils.base import Algorithm


class DAG(Algorithm):
    """Build a Directed Acyclic Graph from an adjacency.

    * Graphs
    * DiGraphs

    Parameters
    ----------
    ordering : str
        A method to sort the nodes.

        * If None, the default order is the index.
        * If ``'degree'``, the nodes are sorted by ascending degree.

    Attributes
    ----------
    indptr_ : np.ndarray
        Pointer index as for CSR format.
    indices_ : np.ndarray
        Indices as for CSR format.
    """
    def __init__(self, ordering: str = None):
        super(DAG, self).__init__()
        self.ordering = ordering
        self.indptr_ = None
        self.indices_ = None

    def fit(self, adjacency: sparse.csr_matrix, sorted_nodes=None):
        """Fit algorithm to the data.

        Parameters
        ----------
        adjacency :
            Adjacency matrix of the graph.
        sorted_nodes : np.ndarray
            An order on the nodes such that the DAG only contains edges (i, j) such that
            ``sorted_nodes[i] < sorted_nodes[j]``.
        """
        indptr = adjacency.indptr.astype(np.int32)
        indices = adjacency.indices.astype(np.int32)

        if sorted_nodes is not None:
            if adjacency.shape[0] != sorted_nodes.shape[0]:
                raise ValueError('Dimensions mismatch between adjacency and sorted_nodes.')
            else:
                sorted_nodes = sorted_nodes.astype(np.int32)
        else:
            if self.ordering is None:
                sorted_nodes = np.arange(adjacency.shape[0]).astype(np.int32)
            elif self.ordering == 'degree':
                degrees = indptr[1:] - indptr[:-1]
                sorted_nodes = np.argsort(degrees).astype(np.int32)
            else:
                raise ValueError('Unknown ordering of nodes.')

        ix = np.zeros(adjacency.shape[0], dtype=np.int32)
        dag_indptr, dag_indices = fit_core(indptr, indices, sorted_nodes, ix)
        self.indptr_ = np.asarray(dag_indptr)
        self.indices_ = np.asarray(dag_indices)

        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Apr 2020
@author: Nathan de Lara <ndelara@enst.fr>
"""

import numpy as np
from scipy import sparse
from scipy.sparse.linalg import LinearOperator


def safe_sparse_dot(a, b):
    """Dot product that handles the sparse matrix case correctly.
    Uses BLAS GEMM as replacement for numpy.dot where possible to avoid unnecessary copies.

    Parameters
    ----------
    a : array, sparse matrix or LinearOperator
    b : array, sparse matrix or LinearOperator
    Returns
    -------
    dot_product : array or sparse matrix
        sparse if ``a`` or ``b`` is sparse.
    """
    if type(a) == np.ndarray:
        return b.T.dot(a.T).T
    if isinstance(a, LinearOperator) and isinstance(b, LinearOperator):
        raise NotImplementedError
    if hasattr(a, 'right_sparse_dot') and type(b) == sparse.csr_matrix:
        if callable(a.right_sparse_dot):
            return a.right_sparse_dot(b)
    if hasattr(b, 'left_sparse_dot') and type(a) == sparse.csr_matrix:
        if callable(b.left_sparse_dot):
            return b.left_sparse_dot(a)
    else:
        return a.dot(b)



# -*- coding: utf-8 -*-
# tests for toy_graphs.py
"""authors:
Thomas Bonald <bonald@enst.fr>
Quentin Lutz <qlutz@enst.fr>"""
import unittest

from sknetwork.data.models import *


class TestModels(unittest.TestCase):

    def test_shape(self):
        n = 10
        for model in [linear_graph, linear_digraph, cyclic_graph, cyclic_digraph, watts_strogatz]:
            adjacency = model(n)
            graph = model(n, metadata=True)
            self.assertEqual(adjacency.shape, (n, n))
            self.assertEqual(graph.adjacency.shape, (n, n))
            if hasattr(graph, 'position'):
                self.assertEqual(graph.position.shape, (n, 2))

        adjacency = erdos_renyi(n)
        self.assertEqual(adjacency.shape, (n, n))

        adjacency = albert_barabasi(n, 2)
        self.assertEqual(adjacency.shape, (n, n))

        n1, n2 = 4, 6
        n = n1 * n2
        adjacency = grid(n1, n2)
        self.assertEqual(adjacency.shape, (n, n))
        graph = grid(n1, n2, metadata=True)
        self.assertEqual(graph.adjacency.shape, (n, n))
        if hasattr(graph, 'position'):
            self.assertEqual(graph.position.shape, (n, 2))

    def test_SBM(self):
        graph = block_model(np.array([4, 5, 6]), np.array([0.5, 0.3, 0.2]), 0.1, metadata=True)
        adjacency = graph.adjacency
        labels = graph.labels
        self.assertEqual(adjacency.shape, (15, 15))
        self.assertEqual(len(labels), 15)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mar, 2020
@author: Nathan de Lara <ndelara@enst.fr>
"""
from typing import Optional

import numpy as np
from scipy import sparse

from sknetwork.classification.base_rank import RankClassifier, RankBiClassifier
from sknetwork.ranking.diffusion import Diffusion, Dirichlet
from sknetwork.utils.check import check_labels


def hot_and_cold_seeds(labels_seeds):
    """Make one-vs-all seed labels from seeds.

    Parameters
    ----------
    labels_seeds :

    Returns
    -------
    seeds_all: list
        Personalization vectors.
    """
    seeds_all = []
    classes, _ = check_labels(labels_seeds)

    for label in classes:
        seeds = -np.ones_like(labels_seeds)
        seeds[labels_seeds == label] = 1
        ix = np.logical_and(labels_seeds != label, labels_seeds >= 0)
        seeds[ix] = 0
        seeds_all.append(seeds)

    return seeds_all


def process_scores(scores: np.ndarray) -> np.ndarray:
    """Post-processing of the score matrix.

    Parameters
    ----------
    scores : np.ndarray
        Matrix of scores, shape number of nodes x number of labels.

    Returns
    -------
    scores: np.ndarray
    """
    scores -= np.mean(scores, axis=0)
    scores = np.exp(scores)
    return scores


class DiffusionClassifier(RankClassifier):
    """Node classification using multiple diffusions.

    * Graphs
    * Digraphs

    Parameters
    ----------
    n_iter : int
        Number of steps of the diffusion in discrete time (must be positive).
    damping_factor : float (optional)
        Damping factor (default value = 1).
    n_jobs :
        If positive, number of parallel jobs allowed (-1 means maximum number).
        If ``None``, no parallel computations are made.

    Attributes
    ----------
    labels_ : np.ndarray
        Label of each node (hard classification).
    membership_ : sparse.csr_matrix
        Membership matrix (soft classification, labels on columns).

    Example
    -------
    >>> from sknetwork.data import karate_club
    >>> diffusion = DiffusionClassifier()
    >>> graph = karate_club(metadata=True)
    >>> adjacency = graph.adjacency
    >>> labels_true = graph.labels
    >>> seeds = {0: labels_true[0], 33: labels_true[33]}
    >>> labels_pred = diffusion.fit_transform(adjacency, seeds)
    >>> np.round(np.mean(labels_pred == labels_true), 2)
    0.94

    References
    ----------
    * de Lara, N., & Bonald, T. (2020).
      `A Consistent Diffusion-Based Algorithm for Semi-Supervised Classification on Graphs.
      <https://arxiv.org/pdf/2008.11944.pdf>`_
      arXiv preprint arXiv:2008.11944.

    * Zhu, X., Lafferty, J., & Rosenfeld, R. (2005). `Semi-supervised learning with graphs
      <http://pages.cs.wisc.edu/~jerryzhu/machineteaching/pub/thesis.pdf>`_
      (Doctoral dissertation, Carnegie Mellon University, language technologies institute, school of computer science).
    """
    def __init__(self, n_iter: int = 10, damping_factor: Optional[float] = None, n_jobs: Optional[int] = None):
        algorithm = Diffusion(n_iter, damping_factor)
        super(DiffusionClassifier, self).__init__(algorithm, n_jobs)
        self._process_scores = process_scores


class BiDiffusionClassifier(DiffusionClassifier, RankBiClassifier):
    """Node classification using multiple diffusions.

    * Bigraphs

    Parameters
    ----------
    n_iter : int
        Number of steps of the diffusion in discrete time (must be positive).
    damping_factor : float (optional)
        Damping factor (default value = 1).
    n_jobs :
        If positive, number of parallel jobs allowed (-1 means maximum number).
        If ``None``, no parallel computations are made.

    Attributes
    ----------
    labels_ : np.ndarray
        Label of each row.
    labels_row_ : np.ndarray
        Label of each row (copy of **labels_**).
    labels_col_ : np.ndarray
        Label of each column.
    membership_ : sparse.csr_matrix
        Membership matrix of rows (soft classification, labels on columns).
    membership_row_ : sparse.csr_matrix
        Membership matrix of rows (copy of **membership_**).
    membership_col_ : sparse.csr_matrix
        Membership matrix of columns.

    Example
    -------
    >>> from sknetwork.classification import BiDiffusionClassifier
    >>> from sknetwork.data import star_wars
    >>> bidiffusion = BiDiffusionClassifier(n_iter=2)
    >>> biadjacency = star_wars()
    >>> seeds = {0: 1, 2: 0}
    >>> bidiffusion.fit_transform(biadjacency, seeds)
    array([1, 1, 0, 0])
    """
    def __init__(self, n_iter: int = 10, damping_factor: Optional[float] = None, n_jobs: Optional[int] = None):
        super(BiDiffusionClassifier, self).__init__(n_iter=n_iter, damping_factor=damping_factor, n_jobs=n_jobs)


class DirichletClassifier(RankClassifier):
    """Node classification using multiple Dirichlet problems.

    * Graphs
    * Digraphs

    Parameters
    ----------
    n_iter : int
        If positive, the solution to the Dirichlet problem is approximated by power iteration for n_iter steps.
        Otherwise, the solution is computed through BiConjugate Stabilized Gradient descent.
    damping_factor : float (optional)
        Damping factor (default value = 1).
    n_jobs :
        If an integer value is given, denotes the number of workers to use (-1 means the maximum number will be used).
        If ``None``, no parallel computations are made.
    verbose :
        Verbose mode.

    Attributes
    ----------
    labels_ : np.ndarray
        Label of each node (hard classification).
    membership_ : sparse.csr_matrix
        Membership matrix (soft classification, labels on columns).

    Example
    -------
    >>> from sknetwork.data import karate_club
    >>> dirichlet = DirichletClassifier()
    >>> graph = karate_club(metadata=True)
    >>> adjacency = graph.adjacency
    >>> labels_true = graph.labels
    >>> seeds = {0: labels_true[0], 33: labels_true[33]}
    >>> labels_pred = dirichlet.fit_transform(adjacency, seeds)
    >>> np.round(np.mean(labels_pred == labels_true), 2)
    0.97

    References
    ----------
    Zhu, X., Lafferty, J., & Rosenfeld, R. (2005). `Semi-supervised learning with graphs
    <http://pages.cs.wisc.edu/~jerryzhu/machineteaching/pub/thesis.pdf>`_
    (Doctoral dissertation, Carnegie Mellon University, language technologies institute, school of computer science).
    """
    def __init__(self, n_iter: int = 10, damping_factor: Optional[float] = None, n_jobs: Optional[int] = None,
                 verbose: bool = False):
        algorithm = Dirichlet(n_iter, damping_factor, verbose)
        super(DirichletClassifier, self).__init__(algorithm, n_jobs, verbose)
        self._process_seeds = hot_and_cold_seeds
        self._process_scores = process_scores


class BiDirichletClassifier(DirichletClassifier, RankBiClassifier):
    """Node classification using multiple diffusions.

    * Bigraphs

    Parameters
    ----------
    n_iter : int
        If positive, the solution to the Dirichlet problem is approximated by power iteration for n_iter steps.
        Otherwise, the solution is computed through BiConjugate Stabilized Gradient descent.
    damping_factor : float (optional)
        Damping factor (default value = 1).
    n_jobs :
        If positive, number of parallel jobs allowed (-1 means maximum number).
        If ``None``, no parallel computations are made.
    verbose :
        Verbose mode.

    Attributes
    ----------
    labels_ : np.ndarray
        Label of each row.
    labels_row_ : np.ndarray
        Label of each row (copy of **labels_**).
    labels_col_ : np.ndarray
        Label of each column.
    membership_ : sparse.csr_matrix
        Membership matrix of rows (soft classification, labels on columns).
    membership_row_ : sparse.csr_matrix
        Membership matrix of rows (copy of **membership_**).
    membership_col_ : sparse.csr_matrix
        Membership matrix of columns.

    Example
    -------
    >>> from sknetwork.data import star_wars
    >>> bidirichlet = BiDirichletClassifier()
    >>> biadjacency = star_wars()
    >>> seeds = {0: 1, 2: 0}
    >>> bidirichlet.fit_transform(biadjacency, seeds)
    array([1, 1, 0, 0])
    """
    def __init__(self, n_iter: int = 10, damping_factor: Optional[float] = None, n_jobs: Optional[int] = None,
                 verbose: bool = False):
        super(BiDirichletClassifier, self).__init__(n_iter=n_iter, damping_factor=damping_factor, verbose=verbose,
                                                    n_jobs=n_jobs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""""tests for search.py"""

import unittest

import numpy as np

from sknetwork.data import cyclic_digraph
from sknetwork.path import breadth_first_search, depth_first_search


class TestSearch(unittest.TestCase):

    def setUp(self) -> None:
        """Load graph for tests."""
        self.adjacency = cyclic_digraph(3).astype(bool)

    def test_bfs(self):
        self.assertTrue((breadth_first_search(
            self.adjacency, 0, return_predecessors=False) == np.array([0, 1, 2])).all())

    def test_dfs(self):
        self.assertTrue((depth_first_search(
            self.adjacency, 0, return_predecessors=False) == np.array([0, 1, 2])).all())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Nov 2, 2018
@author: Nathan de Lara <ndelara@enst.fr>
@author: Quentin Lutz <qlutz@enst.fr>
@author: Thomas Bonald <bonald@enst.fr>
"""
from typing import Union, Optional

import numpy as np
from scipy import sparse

from sknetwork.clustering.base import BaseClustering, BaseBiClustering
from sknetwork.clustering.louvain_core import fit_core
from sknetwork.clustering.postprocess import reindex_labels
from sknetwork.utils.check import check_format, check_random_state, check_probs, check_square
from sknetwork.utils.format import bipartite2directed, directed2undirected, bipartite2undirected
from sknetwork.utils.membership import membership_matrix
from sknetwork.utils.verbose import VerboseMixin


class Louvain(BaseClustering, VerboseMixin):
    """Louvain algorithm for clustering graphs by maximization of modularity.

    * Graphs
    * Digraphs

    Parameters
    ----------
    resolution :
        Resolution parameter.
    modularity : str
        Which objective function to maximize. Can be ``'dugue'``, ``'newman'`` or ``'potts'``.
    tol_optimization :
        Minimum increase in the objective function to enter a new optimization pass.
    tol_aggregation :
        Minimum increase in the objective function to enter a new aggregation pass.
    n_aggregations :
        Maximum number of aggregations.
        A negative value is interpreted as no limit.
    shuffle_nodes :
        Enables node shuffling before optimization.
    sort_clusters :
            If ``True``, sort labels in decreasing order of cluster size.
    return_membership :
            If ``True``, return the membership matrix of nodes to each cluster (soft clustering).
    return_aggregate :
            If ``True``, return the adjacency matrix of the graph between clusters.
    random_state :
        Random number generator or random seed. If None, numpy.random is used.
    verbose :
        Verbose mode.

    Attributes
    ----------
    labels_ : np.ndarray
        Label of each node.
    membership_ : sparse.csr_matrix
        Membership matrix.
    adjacency_ : sparse.csr_matrix
        Adjacency matrix between clusters.

    Example
    -------
    >>> from sknetwork.clustering import Louvain
    >>> from sknetwork.data import karate_club
    >>> louvain = Louvain()
    >>> adjacency = karate_club()
    >>> labels = louvain.fit_transform(adjacency)
    >>> len(set(labels))
    4

    References
    ----------
    * Blondel, V. D., Guillaume, J. L., Lambiotte, R., & Lefebvre, E. (2008).
      `Fast unfolding of communities in large networks.
      <https://arxiv.org/abs/0803.0476>`_
      Journal of statistical mechanics: theory and experiment, 2008

    * Dugué, N., & Perez, A. (2015).
      `Directed Louvain: maximizing modularity in directed networks
      <https://hal.archives-ouvertes.fr/hal-01231784/document>`_
      (Doctoral dissertation, Université d'Orléans).

    * Traag, V. A., Van Dooren, P., & Nesterov, Y. (2011).
      `Narrow scope for resolution-limit-free community detection.
      <https://arxiv.org/pdf/1104.3083.pdf>`_
      Physical Review E, 84(1), 016114.
    """
    def __init__(self, resolution: float = 1, modularity: str = 'dugue', tol_optimization: float = 1e-3,
                 tol_aggregation: float = 1e-3, n_aggregations: int = -1, shuffle_nodes: bool = False,
                 sort_clusters: bool = True, return_membership: bool = True, return_aggregate: bool = True,
                 random_state: Optional[Union[np.random.RandomState, int]] = None, verbose: bool = False):
        super(Louvain, self).__init__(sort_clusters=sort_clusters, return_membership=return_membership,
                                      return_aggregate=return_aggregate)
        VerboseMixin.__init__(self, verbose)

        self.resolution = np.float32(resolution)
        self.modularity = modularity.lower()
        self.tol = np.float32(tol_optimization)
        self.tol_aggregation = tol_aggregation
        self.n_aggregations = n_aggregations
        self.shuffle_nodes = shuffle_nodes
        self.random_state = check_random_state(random_state)

    def _optimize(self, adjacency_norm, probs_ou, probs_in):
        """One local optimization pass of the Louvain algorithm

        Parameters
        ----------
        adjacency_norm :
            the norm of the adjacency
        probs_ou :
            the array of degrees of the adjacency
        probs_in :
            the array of degrees of the transpose of the adjacency

        Returns
        -------
        labels :
            the communities of each node after optimization
        pass_increase :
            the increase in modularity gained after optimization
        """
        node_probs_in = probs_in.astype(np.float32)
        node_probs_ou = probs_ou.astype(np.float32)

        adjacency = 0.5 * directed2undirected(adjacency_norm)

        self_loops = adjacency.diagonal().astype(np.float32)

        indptr: np.ndarray = adjacency.indptr
        indices: np.ndarray = adjacency.indices
        data: np.ndarray = adjacency.data.astype(np.float32)

        return fit_core(self.resolution, self.tol, node_probs_ou, node_probs_in, self_loops, data, indices, indptr)

    @staticmethod
    def _aggregate(adjacency_norm, probs_ou, probs_in, membership: Union[sparse.csr_matrix, np.ndarray]):
        """Aggregate nodes belonging to the same cluster.

        Parameters
        ----------
        adjacency_norm :
            the norm of the adjacency
        probs_ou :
            the array of degrees of the adjacency
        probs_in :
            the array of degrees of the transpose of the adjacency
        membership :
            membership matrix (rows).

        Returns
        -------
        Aggregate graph.
        """
        adjacency_norm = (membership.T.dot(adjacency_norm.dot(membership))).tocsr()
        probs_in = np.array(membership.T.dot(probs_in).T)
        probs_ou = np.array(membership.T.dot(probs_ou).T)
        return adjacency_norm, probs_ou, probs_in

    def fit(self, adjacency: Union[sparse.csr_matrix, np.ndarray]) -> 'Louvain':
        """Fit algorithm to the data.

        Parameters
        ----------
        adjacency :
            Adjacency matrix of the graph.

        Returns
        -------
        self: :class:`Louvain`
        """
        adjacency = check_format(adjacency)
        check_square(adjacency)
        n = adjacency.shape[0]

        if self.modularity == 'potts':
            probs_ou = check_probs('uniform', adjacency)
            probs_in = probs_ou.copy()
        elif self.modularity == 'newman':
            probs_ou = check_probs('degree', adjacency)
            probs_in = probs_ou.copy()
        elif self.modularity == 'dugue':
            probs_ou = check_probs('degree', adjacency)
            probs_in = check_probs('degree', adjacency.T)
        else:
            raise ValueError('Unknown modularity function.')

        nodes = np.arange(n)
        if self.shuffle_nodes:
            nodes = self.random_state.permutation(nodes)
            adjacency = adjacency[nodes, :].tocsc()[:, nodes].tocsr()

        adjacency_clust = adjacency / adjacency.data.sum()

        membership = sparse.identity(n, format='csr')
        increase = True
        count_aggregations = 0
        self.log.print("Starting with", n, "nodes.")
        while increase:
            count_aggregations += 1

            labels_clust, pass_increase = self._optimize(adjacency_clust, probs_ou, probs_in)
            _, labels_clust = np.unique(labels_clust, return_inverse=True)

            if pass_increase <= self.tol_aggregation:
                increase = False
            else:
                membership_clust = membership_matrix(labels_clust)
                membership = membership.dot(membership_clust)
                adjacency_clust, probs_ou, probs_in = self._aggregate(adjacency_clust, probs_ou, probs_in,
                                                                      membership_clust)

                n = adjacency_clust.shape[0]
                if n == 1:
                    break
            self.log.print("Aggregation", count_aggregations, "completed with", n, "clusters and ",
                           pass_increase, "increment.")
            if count_aggregations == self.n_aggregations:
                break

        if self.sort_clusters:
            labels = reindex_labels(membership.indices)
        else:
            labels = membership.indices
        if self.shuffle_nodes:
            reverse = np.empty(nodes.size, nodes.dtype)
            reverse[nodes] = np.arange(nodes.size)
            labels = labels[reverse]

        self.labels_ = labels
        self._secondary_outputs(adjacency)

        return self


class BiLouvain(Louvain, BaseBiClustering):
    """BiLouvain algorithm for the clustering of bipartite graphs.

    * Bigraphs

    Parameters
    ----------
    resolution :
        Resolution parameter.
    modularity : str
        Which objective function to maximize. Can be ``'dugue'``, ``'newman'`` or ``'potts'``.
    tol_optimization :
        Minimum increase in the objective function to enter a new optimization pass.
    tol_aggregation :
        Minimum increase in the objective function to enter a new aggregation pass.
    n_aggregations :
        Maximum number of aggregations.
        A negative value is interpreted as no limit.
    shuffle_nodes :
        Enables node shuffling before optimization.
    sort_clusters :
            If ``True``, sort labels in decreasing order of cluster size.
    return_membership :
            If ``True``, return the membership matrix of nodes to each cluster (soft clustering).
    return_aggregate :
        If ``True``, return the biadjacency matrix of the graph between clusters.
    random_state :
        Random number generator or random seed. If None, numpy.random is used.
    verbose :
        Verbose mode.

    Attributes
    ----------
    labels_ : np.ndarray
        Labels of the rows.
    labels_row_ : np.ndarray
        Labels of the rows (copy of **labels_**).
    labels_col_ : np.ndarray
        Labels of the columns.
    membership_row_ : sparse.csr_matrix
        Membership matrix of the rows (copy of **membership_**).
    membership_col_ : sparse.csr_matrix
        Membership matrix of the columns. Only valid if **cluster_both** = `True`.
    biadjacency_ : sparse.csr_matrix
        Biadjacency matrix of the aggregate graph between clusters.

    Example
    -------
    >>> from sknetwork.clustering import BiLouvain
    >>> from sknetwork.data import movie_actor
    >>> bilouvain = BiLouvain()
    >>> biadjacency = movie_actor()
    >>> labels = bilouvain.fit_transform(biadjacency)
    >>> len(labels)
    15
    """
    def __init__(self, resolution: float = 1, modularity: str = 'dugue', tol_optimization: float = 1e-3,
                 tol_aggregation: float = 1e-3, n_aggregations: int = -1, shuffle_nodes: bool = False,
                 sort_clusters: bool = True, return_membership: bool = True, return_aggregate: bool = True,
                 random_state: Optional[Union[np.random.RandomState, int]] = None, verbose: bool = False):
        super(BiLouvain, self).__init__(sort_clusters=sort_clusters, return_membership=return_membership,
                                        return_aggregate=return_aggregate, resolution=resolution, modularity=modularity,
                                        tol_optimization=tol_optimization, verbose=verbose,
                                        tol_aggregation=tol_aggregation, n_aggregations=n_aggregations,
                                        shuffle_nodes=shuffle_nodes, random_state=random_state)

    def fit(self, biadjacency: Union[sparse.csr_matrix, np.ndarray]) -> 'BiLouvain':
        """Apply the Louvain algorithm to the corresponding directed graph, with adjacency matrix:

        :math:`A  = \\begin{bmatrix} 0 & B \\\\ 0 & 0 \\end{bmatrix}`

        where :math:`B` is the input (biadjacency matrix).

        Parameters
        ----------
        biadjacency:
            Biadjacency matrix of the graph.

        Returns
        -------
        self: :class:`BiLouvain`
        """
        louvain = Louvain(resolution=self.resolution, modularity=self.modularity, tol_aggregation=self.tol_aggregation,
                          n_aggregations=self.n_aggregations, shuffle_nodes=self.shuffle_nodes,
                          sort_clusters=self.sort_clusters, return_membership=self.return_membership,
                          return_aggregate=False, random_state=self.random_state, verbose=self.log.verbose)
        biadjacency = check_format(biadjacency)
        n_row, _ = biadjacency.shape

        if self.modularity == 'dugue':
            adjacency = bipartite2directed(biadjacency)
        else:
            adjacency = bipartite2undirected(biadjacency)
        louvain.fit(adjacency)

        self.labels_ = louvain.labels_
        self._split_vars(n_row)
        self._secondary_outputs(biadjacency)

        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Tests for k-cliques count"""
import unittest

from scipy.special import comb

from sknetwork.data.test_graphs import *
from sknetwork.topology import Cliques


class TestCliqueListing(unittest.TestCase):

    def test_empty(self):
        adjacency = test_graph_empty()
        self.assertEqual(Cliques(2).fit_transform(adjacency), 0)
        cliques = Cliques(1)
        self.assertRaises(ValueError, cliques.fit_transform, adjacency)

    def test_disconnected(self):
        adjacency = test_graph_disconnect()
        self.assertEqual(Cliques(3).fit_transform(adjacency), 1)

    def test_cliques(self):
        adjacency = test_graph_clique()
        n = adjacency.shape[0]
        self.assertEqual(Cliques(3).fit_transform(adjacency), comb(n, 3, exact=True))
        self.assertEqual(Cliques(4).fit_transform(adjacency), comb(n, 4, exact=True))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on October 2019
@author: Nathan de Lara <ndelara@enst.fr>
"""

import unittest

from sknetwork.clustering import BiKMeans, KMeans
from sknetwork.data.test_graphs import *
from sknetwork.embedding import GSVD, SVD, BiSpectral


class TestKMeans(unittest.TestCase):

    def setUp(self):
        self.kmeans = KMeans(3, GSVD(2))
        self.bikmeans = BiKMeans(3, GSVD(2))
        self.kmeans_options = KMeans(4, SVD(3), sort_clusters=False)
        self.bikmeans_options = BiKMeans(4, BiSpectral(3), co_cluster=True, sort_clusters=False)

    def test_undirected(self):
        for adjacency in [test_graph(), test_graph_disconnect()]:
            n = adjacency.shape[0]
            labels = self.kmeans.fit_transform(adjacency)
            self.assertEqual(len(set(labels)), 3)
            self.assertEqual(self.kmeans.membership_.shape, (n, 3))
            self.assertEqual(self.kmeans.adjacency_.shape, (3, 3))
            labels = self.kmeans_options.fit_transform(adjacency)
            self.assertEqual(len(set(labels)), 4)

    def test_directed(self):
        adjacency = test_digraph()
        n = adjacency.shape[0]
        labels = self.kmeans.fit_transform(adjacency)
        self.assertEqual(len(set(labels)), 3)
        self.assertEqual(self.kmeans.membership_.shape, (n, 3))
        self.assertEqual(self.kmeans.adjacency_.shape, (3, 3))
        labels = self.kmeans_options.fit_transform(adjacency)
        self.assertEqual(len(set(labels)), 4)

    def test_bipartite(self):
        for biadjacency in [test_bigraph(), test_bigraph_disconnect()]:
            n_row, n_col = biadjacency.shape
            self.bikmeans.fit(biadjacency)
            self.assertEqual(len(set(self.bikmeans.labels_row_)), 3)
            self.assertEqual(self.bikmeans.membership_.shape, (n_row, 3))
            self.assertEqual(self.bikmeans.membership_row_.shape, (n_row, 3))
            self.assertEqual(self.bikmeans.biadjacency_.shape, (3, n_col))
            self.bikmeans_options.fit(biadjacency)
            labels = np.hstack((self.bikmeans_options.labels_row_, self.bikmeans_options.labels_col_))
            self.assertEqual(len(set(labels)), 4)
            self.assertEqual(self.bikmeans_options.membership_.shape, (n_row, 4))
            self.assertEqual(self.bikmeans_options.membership_row_.shape, (n_row, 4))
            self.assertEqual(self.bikmeans_options.membership_col_.shape, (n_col, 4))
            self.assertEqual(self.bikmeans_options.biadjacency_.shape, (4, 4))




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Apr 2020
@author: Nathan de Lara <ndelara@enst.fr>
"""
from typing import Union

import numpy as np
from scipy import sparse
from scipy.sparse.linalg import eigs, LinearOperator, bicgstab

from sknetwork.linalg.diteration import diffusion
from sknetwork.linalg.normalization import normalize
from sknetwork.linalg.polynome import Polynome


class RandomSurferOperator(LinearOperator):
    """Random surfer as a LinearOperator

    Parameters
    ----------
    adjacency :
        Adjacency matrix of the graph as a CSR or a LinearOperator.
    damping_factor : float
        Probability to continue the random walk.
    seeds :
        Probability vector for seeds.

    Attributes
    ----------
    a : sparse.csr_matrix
        Scaled transposed transition matrix.
    b : np.ndarray
        Scaled restart probability vector.
    """
    def __init__(self, adjacency: Union[sparse.csr_matrix, LinearOperator], seeds: np.ndarray, damping_factor):
        super(RandomSurferOperator, self).__init__(shape=adjacency.shape, dtype=float)

        n = adjacency.shape[0]
        out_degrees = adjacency.dot(np.ones(n)).astype(bool)

        if hasattr(adjacency, 'left_sparse_dot'):
            self.a = damping_factor * normalize(adjacency).T
        else:
            self.a = (damping_factor * normalize(adjacency)).T.tocsr()
        self.b = (np.ones(n) - damping_factor * out_degrees) * seeds

    def _matvec(self, x: np.ndarray):
        return self.a.dot(x) + self.b * x.sum()


def get_pagerank(adjacency: Union[sparse.csr_matrix, LinearOperator], seeds: np.ndarray, damping_factor: float,
                 n_iter: int, tol: float = 1e-6, solver: str = 'piteration') -> np.ndarray:
    """Solve the Pagerank problem. Formally,

    :math:`x = \\alpha Px + (1-\\alpha)y`,

    where :math:`P = (D^{-1}A)^T` is the transition matrix and :math:`y` is the personalization probability vector.

    Parameters
    ----------
    adjacency : sparse.csr_matrix
        Adjacency matrix of the graph.
    seeds : np.ndarray
        Personalization array. Must be a valid probability vector.
    damping_factor : float
        Probability to continue the random walk.
    n_iter : int
        Number of iterations for some of the solvers such as ``'piteration'`` or ``'diteration'``.
    tol : float
        Tolerance for the convergence of some solvers such as ``'bicgstab'`` or ``'lanczos'``.
    solver : :obj:`str`
        Which solver to use: ``'piteration'``, ``'diteration'``, ``'bicgstab'``, ``'lanczos'``, `̀'RH'``.

    Returns
    -------
    pagerank : np.ndarray
        Probability vector.

    Examples
    --------
    >>> from sknetwork.data import house
    >>> adjacency = house()
    >>> seeds = np.array([1, 0, 0, 0, 0])
    >>> scores = get_pagerank(adjacency, seeds, damping_factor=0.85, n_iter=10)
    >>> np.round(scores, 2)
    array([0.29, 0.24, 0.12, 0.12, 0.24])


    References
    ----------
    * Hong, D. (2012). `Optimized on-line computation of pagerank algorithm.
      <https://arxiv.org/pdf/1202.6158.pdf>`_
      arXiv preprint arXiv:1202.6158.
    * Van der Vorst, H. A. (1992). `Bi-CGSTAB:
      <https://en.wikipedia.org/wiki/Biconjugate_gradient_stabilized_method>`_
      A fast and smoothly converging variant of Bi-CG for the solution of nonsymmetric linear systems.
      SIAM Journal on scientific and Statistical Computing, 13(2), 631-644.
    * Lanczos, C. (1950).
      `An iteration method for the solution of the eigenvalue problem of linear differential and integral operators.
      <http://www.cs.umd.edu/~oleary/lanczos1950.pdf>`_
      Los Angeles, CA: United States Governm. Press Office.
    """
    n = adjacency.shape[0]

    if solver == 'diteration':
        if not isinstance(adjacency, sparse.csr_matrix):
            raise ValueError('D-iteration is not compatible with linear operators.')
        adjacency = normalize(adjacency, p=1)
        indptr = adjacency.indptr.astype(np.int32)
        indices = adjacency.indices.astype(np.int32)
        data = adjacency.data.astype(np.float32)
        damping_factor = np.float32(damping_factor)
        n_iter = np.int32(n_iter)
        tol = np.float32(tol)

        scores = np.zeros(n, dtype=np.float32)
        fluid = (1 - damping_factor) * seeds.astype(np.float32)
        diffusion(indptr, indices, data, scores, fluid, damping_factor, n_iter, tol)

    elif solver == 'RH':
        coeffs = np.ones(n_iter+1)
        polynome = Polynome(damping_factor * normalize(adjacency, p=1).T.tocsr(), coeffs)
        scores = polynome.dot(seeds)

    else:
        rso = RandomSurferOperator(adjacency, seeds, damping_factor)
        v0 = rso.b
        if solver == 'bicgstab':
            scores, info = bicgstab(sparse.eye(n, format='csr') - rso.a, rso.b, atol=tol, x0=v0)
        elif solver == 'lanczos':
            # noinspection PyTypeChecker
            _, scores = sparse.linalg.eigs(rso, k=1, tol=tol, v0=v0)
            scores = abs(scores.flatten().real)
        elif solver == 'piteration':
            scores = v0
            for i in range(n_iter):
                scores_ = rso.dot(scores)
                scores_ /= scores_.sum()
                if np.linalg.norm(scores - scores_, ord=1) < tol:
                    break
                else:
                    scores = scores_
        else:
            raise ValueError('Unknown solver.')

    return scores / scores.sum()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on July 2, 2020
@author: Pierre Pebereau <pierre.pebereau@telecom-paris.fr>
@author: Alexis Barreaux <alexis.barreaux@telecom-paris.fr>
"""
from typing import Union

import numpy as np
from scipy import sparse
from sknetwork.topology.weisfeiler_lehman_core import weisfeiler_lehman_coloring

from sknetwork.utils.base import Algorithm


class WeisfeilerLehman(Algorithm):
    """Weisfeiler-Lehman algorithm for coloring/labeling graphs in order to check similarity.

    Parameters
    ----------
    max_iter : int
        Maximum number of iterations. Negative value  means until convergence.

    Attributes
    ----------
    labels_ : np.ndarray
        Label of each node.

    Example
    -------
    >>> from sknetwork.topology import WeisfeilerLehman
    >>> from sknetwork.data import house
    >>> weisfeiler_lehman = WeisfeilerLehman()
    >>> adjacency = house()
    >>> labels = weisfeiler_lehman.fit_transform(adjacency)
    >>> labels
    array([0, 2, 1, 1, 2], dtype=int32)

    References
    ----------
    * Douglas, B. L. (2011).
      `The Weisfeiler-Lehman Method and Graph Isomorphism Testing.
      <https://arxiv.org/pdf/1101.5211.pdf>`_

    * Shervashidze, N., Schweitzer, P., van Leeuwen, E. J., Melhorn, K., Borgwardt, K. M. (2011)
      `Weisfeiler-Lehman graph kernels.
      <http://www.jmlr.org/papers/volume12/shervashidze11a/shervashidze11a.pdf>`_
      Journal of Machine Learning Research 12, 2011.
    """
    def __init__(self, max_iter: int = -1):
        super(WeisfeilerLehman, self).__init__()
        self.max_iter = max_iter
        self.labels_ = None

    def fit(self, adjacency: Union[sparse.csr_matrix, np.ndarray]) -> 'WeisfeilerLehman':
        """Fit algorithm to the data.

        Parameters
        ----------
        adjacency : Union[sparse.csr_matrix, np.ndarray]
            Adjacency matrix of the graph.

        Returns
        -------
        self: :class:`WeisfeilerLehman`
        """
        n: int = adjacency.shape[0]
        if self.max_iter < 0 or self.max_iter > n:
            max_iter = np.int32(n)
        else:
            max_iter = np.int32(self.max_iter)

        labels = np.zeros(n, dtype=np.int32)
        powers = (-np.pi / 3.15) ** np.arange(n, dtype=np.double)
        indptr = adjacency.indptr.astype(np.int32)
        indices = adjacency.indices.astype(np.int32)

        labels, _ = weisfeiler_lehman_coloring(indptr, indices, labels, powers, max_iter)
        self.labels_ = np.asarray(labels).astype(np.int32)
        return self

    def fit_transform(self, adjacency: Union[sparse.csr_matrix, np.ndarray]) -> np.ndarray:
        """Fit algorithm to the data and return the labels. Same parameters as the ``fit`` method.

        Returns
        -------
        labels : np.ndarray
            Labels.
        """
        self.fit(adjacency)
        return self.labels_


def are_isomorphic(adjacency1: sparse.csr_matrix,
                   adjacency2: sparse.csr_matrix, max_iter: int = -1) -> bool:
    """Weisfeiler-Lehman isomorphism test. If the test is False, the graphs cannot be isomorphic,
    otherwise, they might be.

    Parameters
    -----------
    adjacency1 :
        First adjacency matrix.
    adjacency2 :
        Second adjacency matrix.
    max_iter : int
        Maximum number of coloring iterations. Negative value means until convergence.

    Returns
    -------
    test_result : bool

    Example
    -------
    >>> from sknetwork.topology import are_isomorphic
    >>> from sknetwork.data import house
    >>> adjacency_1 = house()
    >>> adjacency_2 = house()
    >>> are_isomorphic(adjacency_1, adjacency_2)
    True

    References
    ----------
    * Douglas, B. L. (2011).
      `The Weisfeiler-Lehman Method and Graph Isomorphism Testing.
      <https://arxiv.org/pdf/1101.5211.pdf>`_

    * Shervashidze, N., Schweitzer, P., van Leeuwen, E. J., Melhorn, K., Borgwardt, K. M. (2011)
      `Weisfeiler-Lehman graph kernels.
      <http://www.jmlr.org/papers/volume12/shervashidze11a/shervashidze11a.pdf>`_
      Journal of Machine Learning Research 12, 2011.
    """
    if (adjacency1.shape != adjacency2.shape) or (adjacency1.nnz != adjacency2.nnz):
        return False

    n = adjacency1.shape[0]

    if max_iter < 0 or max_iter > n:
        max_iter = n

    indptr1 = adjacency1.indptr.astype(np.int32)
    indptr2 = adjacency2.indptr.astype(np.int32)
    indices1 = adjacency1.indices.astype(np.int32)
    indices2 = adjacency2.indices.astype(np.int32)

    labels_1 = np.zeros(n, dtype=np.int32)
    labels_2 = np.zeros(n, dtype=np.int32)

    powers = (- np.pi / 3.15) ** np.arange(n, dtype=np.double)

    iteration = 0
    has_changed_1, has_changed_2 = True, True
    while iteration < max_iter and (has_changed_1 or has_changed_2):
        labels_1, has_changed_1 = weisfeiler_lehman_coloring(indptr1, indices1, labels_1, powers, max_iter=1)
        labels_2, has_changed_2 = weisfeiler_lehman_coloring(indptr2, indices2, labels_2, powers, max_iter=1)

        colors_1, counts_1 = np.unique(np.asarray(labels_1), return_counts=True)
        colors_2, counts_2 = np.unique(np.asarray(labels_2), return_counts=True)

        if (colors_1.shape != colors_2.shape) or (counts_1 != counts_2).any():
            return False

        iteration += 1

    return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mar, 2020
@author: Nathan de Lara <ndelara@enst.fr>
"""
from typing import Optional

import numpy as np
from scipy import sparse

from sknetwork.classification.base_rank import RankClassifier, RankBiClassifier
from sknetwork.ranking.pagerank import PageRank


class PageRankClassifier(RankClassifier):
    """Node classification by multiple personalized PageRanks.

    * Graphs
    * Digraphs

    Parameters
    ----------
    damping_factor:
        Probability to continue the random walk.
    solver : :obj:`str`
        Which solver to use: 'piteration', 'diteration', 'bicgstab', 'lanczos'.
    n_iter : int
        Number of iterations for some of the solvers such as ``'piteration'`` or ``'diteration'``.
    tol : float
        Tolerance for the convergence of some solvers such as ``'bicgstab'`` or ``'lanczos'``.

    Attributes
    ----------
    labels_ : np.ndarray
        Label of each node (hard classification).
    membership_ : sparse.csr_matrix
        Membership matrix (soft classification, columns = labels).

    Example
    -------
    >>> from sknetwork.classification import PageRankClassifier
    >>> from sknetwork.data import karate_club
    >>> pagerank = PageRankClassifier()
    >>> graph = karate_club(metadata=True)
    >>> adjacency = graph.adjacency
    >>> labels_true = graph.labels
    >>> seeds = {0: labels_true[0], 33: labels_true[33]}
    >>> labels_pred = pagerank.fit_transform(adjacency, seeds)
    >>> np.round(np.mean(labels_pred == labels_true), 2)
    0.97

    References
    ----------
    Lin, F., & Cohen, W. W. (2010). `Semi-supervised classification of network data using very few labels.
    <https://lti.cs.cmu.edu/sites/default/files/research/reports/2009/cmulti09017.pdf>`_
    In IEEE International Conference on Advances in Social Networks Analysis and Mining.
    """
    def __init__(self, damping_factor: float = 0.85, solver: str = 'piteration', n_iter: int = 10, tol: float = 0.,
                 n_jobs: Optional[int] = None, verbose: bool = False):
        algorithm = PageRank(damping_factor, solver, n_iter, tol)
        super(PageRankClassifier, self).__init__(algorithm, n_jobs, verbose)


class BiPageRankClassifier(PageRankClassifier, RankBiClassifier):
    """Node classification for bipartite graphs by multiple personalized PageRanks .

    * Bigraphs

    Parameters
    ----------
    damping_factor:
        Probability to continue the random walk.
    solver : :obj:`str`
        Which solver to use: 'piteration', 'diteration', 'bicgstab', 'lanczos'.
    n_iter : int
        Number of iterations for some of the solvers such as ``'piteration'`` or ``'diteration'``.
    tol : float
        Tolerance for the convergence of some solvers such as ``'bicgstab'`` or ``'lanczos'``.

    Attributes
    ----------
    labels_ : np.ndarray
        Label of each row.
    labels_row_ : np.ndarray
        Label of each row (copy of **labels_**).
    labels_col_ : np.ndarray
        Label of each column.
    membership_ : sparse.csr_matrix
        Membership matrix of rows (soft classification, labels on columns).
    membership_row_ : sparse.csr_matrix
        Membership matrix of rows (copy of **membership_**).
    membership_col_ : sparse.csr_matrix
        Membership matrix of columns.

    Example
    -------
    >>> from sknetwork.classification import BiPageRankClassifier
    >>> from sknetwork.data import star_wars
    >>> bipagerank = BiPageRankClassifier()
    >>> biadjacency = star_wars()
    >>> seeds = {0: 1, 2: 0}
    >>> bipagerank.fit_transform(biadjacency, seeds)
    array([1, 1, 0, 0])
    """

    def __init__(self, damping_factor: float = 0.85, solver: str = 'piteration', n_iter: int = 10, tol: float = 0.,
                 n_jobs: Optional[int] = None, verbose: bool = False):
        super(BiPageRankClassifier, self).__init__(damping_factor=damping_factor, solver=solver, n_iter=n_iter, tol=tol,
                                                   n_jobs=n_jobs, verbose=verbose)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on March 2020
@author: Quentin Lutz <qlutz@enst.fr>
@author: Thomas Bonald <tbonald@enst.fr>
"""
from typing import Optional, Union

import numpy as np
from scipy import sparse

from sknetwork.clustering.louvain import Louvain
from sknetwork.hierarchy.base import BaseHierarchy, BaseBiHierarchy
from sknetwork.hierarchy.postprocess import get_dendrogram, reorder_dendrogram
from sknetwork.utils.check import check_format, check_square
from sknetwork.utils.format import bipartite2undirected


class LouvainHierarchy(BaseHierarchy):
    """Hierarchical clustering by successive instances of Louvain (top-down).

    * Graphs
    * Digraphs

    Parameters
    ----------
    depth :
        Depth of the tree.
        A negative value is interpreted as no limit (return a tree of maximum depth).
    resolution :
        Resolution parameter.
    tol_optimization :
        Minimum increase in the objective function to enter a new optimization pass.
    tol_aggregation :
        Minimum increase in the objective function to enter a new aggregation pass.
    n_aggregations :
        Maximum number of aggregations.
        A negative value is interpreted as no limit.
    shuffle_nodes :
        Enables node shuffling before optimization.
    random_state :
        Random number generator or random seed. If ``None``, numpy.random is used.
    verbose :
        Verbose mode.

    Attributes
    ----------
    dendrogram_ : np.ndarray
        Dendrogram.

    Example
    -------
    >>> from sknetwork.hierarchy import LouvainHierarchy
    >>> from sknetwork.data import house
    >>> louvain = LouvainHierarchy()
    >>> adjacency = house()
    >>> louvain.fit_transform(adjacency)
    array([[3., 2., 0., 2.],
           [4., 1., 0., 2.],
           [6., 0., 0., 3.],
           [5., 7., 1., 5.]])

    Notes
    -----
    Each row of the dendrogram = merge nodes, distance, size of cluster.

    See Also
    --------
    scipy.cluster.hierarchy.dendrogram
    """

    def __init__(self, depth: int = 3, resolution: float = 1, tol_optimization: float = 1e-3,
                 tol_aggregation: float = 1e-3, n_aggregations: int = -1, shuffle_nodes: bool = False,
                 random_state: Optional[Union[np.random.RandomState, int]] = None, verbose: bool = False):
        super(LouvainHierarchy, self).__init__()

        self.depth = depth
        self._clustering_method = Louvain(resolution=resolution, tol_optimization=tol_optimization,
                                          tol_aggregation=tol_aggregation, n_aggregations=n_aggregations,
                                          shuffle_nodes=shuffle_nodes, random_state=random_state, verbose=verbose)

    def _recursive_louvain(self, adjacency: Union[sparse.csr_matrix, np.ndarray], depth: int,
                           nodes: Optional[np.ndarray] = None):
        """Recursive function for fit.

        Parameters
        ----------
        adjacency :
            Adjacency matrix of the graph.
        depth :
            Depth of the recursion.
        nodes :
            The current nodes index in the original graph.

        Returns
        -------
        tree: :class:`Tree`
        """
        n = adjacency.shape[0]
        if nodes is None:
            nodes = np.arange(n)

        if adjacency.nnz and depth:
            labels = self._clustering_method.fit_transform(adjacency)
        else:
            labels = np.zeros(n)

        clusters = np.unique(labels)

        result = []
        if len(clusters) == 1:
            if len(nodes) > 1:
                return [[node] for node in nodes]
            else:
                return [nodes[0]]
        else:
            for cluster in clusters:
                mask = (labels == cluster)
                nodes_cluster = nodes[mask]
                adjacency_cluster = adjacency[mask, :][:, mask]
                result.append(self._recursive_louvain(adjacency_cluster, depth - 1, nodes_cluster))
            return result

    def fit(self, adjacency: Union[sparse.csr_matrix, np.ndarray]) -> 'LouvainHierarchy':
        """Fit algorithm to data.

        Parameters
        ----------
        adjacency :
            Adjacency matrix of the graph.

        Returns
        -------
        self: :class:`LouvainHierarchy`
        """
        adjacency = check_format(adjacency)
        check_square(adjacency)

        tree = self._recursive_louvain(adjacency, self.depth)
        dendrogram, _ = get_dendrogram(tree)
        dendrogram = np.array(dendrogram)
        dendrogram[:, 2] -= min(dendrogram[:, 2])

        self.dendrogram_ = reorder_dendrogram(dendrogram)

        return self


class BiLouvainHierarchy(LouvainHierarchy, BaseBiHierarchy):
    """Hierarchical clustering of bipartite graphs by successive instances of Louvain (top-down).

    * Bigraphs

    Parameters
    ----------
    depth :
        Depth of the tree.
        A negative value is interpreted as no limit (return a tree of maximum depth).
    resolution :
        Resolution parameter.
    tol_optimization :
        Minimum increase in the objective function to enter a new optimization pass.
    tol_aggregation :
        Minimum increase in the objective function to enter a new aggregation pass.
    n_aggregations :
        Maximum number of aggregations.
        A negative value is interpreted as no limit.
    shuffle_nodes :
        Enables node shuffling before optimization.
    random_state :
        Random number generator or random seed. If None, numpy.random is used.
    verbose :
        Verbose mode.

    Attributes
    ----------
    dendrogram_ :
        Dendrogram for the rows.
    dendrogram_row_ :
        Dendrogram for the rows (copy of **dendrogram_**).
    dendrogram_col_ :
        Dendrogram for the columns.
    dendrogram_full_ :
        Dendrogram for both rows and columns, indexed in this order.

    Examples
    --------
    >>> from sknetwork.hierarchy import BiLouvainHierarchy
    >>> from sknetwork.data import star_wars
    >>> bilouvain = BiLouvainHierarchy()
    >>> biadjacency = star_wars()
    >>> bilouvain.fit_transform(biadjacency)
    array([[3., 2., 1., 2.],
           [1., 0., 1., 2.],
           [5., 4., 2., 4.]])

    Notes
    -----
    Each row of the dendrogram = :math:`i, j`, height, size of cluster.

    See Also
    --------
    scipy.cluster.hierarchy.linkage
    """
    def __init__(self, **kwargs):
        super(BiLouvainHierarchy, self).__init__()

        self.louvain_hierarchy = LouvainHierarchy(**kwargs)

    def fit(self, biadjacency: Union[sparse.csr_matrix, np.ndarray]) -> 'BiLouvainHierarchy':
        """Applies Louvain hierarchical clustering to

        :math:`A  = \\begin{bmatrix} 0 & B \\\\ B^T & 0 \\end{bmatrix}`

        where :math:`B` is the biadjacency matrix of the graphs.

        Parameters
        ----------
        biadjacency:
            Biadjacency matrix of the graph.

        Returns
        -------
        self: :class:`BiLouvainHierarchy`
        """
        biadjacency = check_format(biadjacency)
        adjacency = bipartite2undirected(biadjacency)

        self.dendrogram_ = self.louvain_hierarchy.fit_transform(adjacency)
        self._split_vars(biadjacency.shape)

        return self



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Nov 29, 2018
@author: Quentin Lutz <qlutz@enst.fr>
@author: Nathan de Lara <ndelara@enst.fr>
@author: Thomas Bonald <tbonald@enst.fr>
"""

import numpy as np
from scipy import sparse

from sknetwork.utils import directed2undirected


def test_digraph():
    """Simple directed graph, used for testing.
    10 nodes, 10 edges
    """
    row = np.array([0, 0, 1, 3, 4, 6, 6, 6, 7, 8, 9])
    col = np.array([1, 4, 3, 2, 5, 4, 5, 7, 9, 9, 9])
    data = np.array([1, 1, 2.5, 1, 2, 2, 1, 2, 2, 1.5, 2])
    return sparse.csr_matrix((data, (row, col)), shape=(10, 10))


def test_graph():
    """Simple undirected graph, used for testing.
    10 nodes, 10 edges.
    """
    return directed2undirected(test_digraph(), weighted=True)


def test_bigraph():
    """Simple bipartite graph, used for testing.
    6 + 8 nodes, 9 edges.
    """
    row = np.array([0, 1, 1, 2, 2, 3, 4, 5, 5])
    col = np.array([1, 2, 3, 1, 0, 4, 7, 7, 6])
    data = np.array([1, 2.5, 1, 2, 2, 1.5, 1, 0, 3])
    return sparse.csr_matrix((data, (row, col)), shape=(6, 8))


def test_graph_disconnect():
    """Simple disconnected undirected graph, used for testing.
    10 nodes, 10 edges.
    """
    row = np.array([1, 2, 3, 4, 6, 6, 6, 7, 8, 9])
    col = np.array([1, 3, 2, 5, 4, 5, 7, 9, 9, 9])
    data = np.array([1, 2.5, 1, 2, 2, 1, 2, 2, 1.5, 2])
    adjacency = sparse.csr_matrix((data, (row, col)), shape=(10, 10))
    return directed2undirected(adjacency)


def test_bigraph_disconnect():
    """Simple disconnected bipartite graph, used for testing.
    6 + 8 nodes, 9 edges.
    """
    row = np.array([1, 1, 1, 2, 2, 3, 5, 4, 5])
    col = np.array([1, 2, 3, 1, 3, 4, 7, 7, 6])
    data = np.array([1, 2.5, 1, 2, 2, 1.5, 3, 0, 1])
    return sparse.csr_matrix((data, (row, col)), shape=(6, 8))


def test_graph_bool():
    """Simple undirected graph with boolean entries, used for testing (10 nodes, 10 edges)."""
    adjacency = test_graph()
    adjacency.data = adjacency.data.astype(bool)
    return adjacency


def test_graph_clique():
    """Clique graph, used for testing (10 nodes, 45 edges).
    """
    n = 10
    adjacency = sparse.csr_matrix(np.ones((n, n), dtype=bool))
    adjacency.setdiag(0)
    adjacency.eliminate_zeros()
    return adjacency


def test_graph_empty():
    """Empty graph, used for testing (10 nodes, 0 edges).
    """
    return sparse.csr_matrix((10, 10), dtype=bool)
